# (c) 2015-2022 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import functools
import inspect
import warnings


class _Deprecated(object):
    """
    Flags a class or function as deprecated

    Parameters
    ----------
    version : str
        The HTMD version at which the class or function was deprecated.
    newname : str
        The replacement function, with complete Namespace.

    Examples
    --------
    >>> @_Deprecated('0.0.0')
    ... class ExampleClass:
    ...     @_Deprecated('0.0.1')
    ...     def example_method(self):
    ...         pass
    ...     pass
    >>> ExampleClass.__doc__
    '\\n    .. warning:: Deprecated since version 0.0.0. \\n                '
    >>> ExampleClass().example_method.__doc__
    '\\n        .. warning:: Deprecated since version 0.0.1. \\n                '
    >>> ExampleClass() # doctest:+ELLIPSIS
    <__main__.ExampleClass object ...>
    >>> ExampleClass().example_method # doctest:+ELLIPSIS
    <bound method ExampleClass.example_method of <__main__.ExampleClass object ...>>
    """

    def __init__(self, version, newname=None):
        self.version = version
        self.newname = newname

    def __call__(self, wrapped):

        if inspect.isfunction(wrapped):
            objtype = "function/method"
            docprefix = """
        .. warning:: Deprecated since version {}. {}
                """.format(
                self.version,
                "Use :func:`{}` instead.".format(self.newname) if self.newname else "",
            )
        elif inspect.isclass(wrapped):
            objtype = "class"
            docprefix = """
    .. warning:: Deprecated since version {}. {}
                """.format(
                self.version,
                "Use :func:`{}` instead.".format(self.newname) if self.newname else "",
            )
        else:
            raise TypeError(type(wrapped))

        message = "The `{}` {} has been deprecated " "since version {}. {}".format(
            wrapped.__name__,
            objtype,
            self.version,
            "Use `{}` instead.".format(self.newname) if self.newname else "",
        )

        wrappingclass = isinstance(wrapped, type)
        if wrappingclass:
            wrappedfunc = wrapped.__init__
        else:
            wrappedfunc = wrapped

        wrapped.__doc__ = docprefix + (wrapped.__doc__ or "")

        @functools.wraps(wrappedfunc)
        def new_func(*args, **kwargs):
            warnings.warn(message, category=DeprecationWarning, stacklevel=2)
            return wrappedfunc(*args, **kwargs)

        if wrappingclass:
            wrapped.__init__ = new_func
            return wrapped
        else:
            return new_func


if __name__ == "__main__":
    import doctest

    doctest.testmod()



# (c) 2015-2022 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import os
import logging

logger = logging.getLogger(__file__)

_config = {
    "njobs": 1,
    "configfile": os.getenv("HTMD_CONFIG") if os.getenv("HTMD_CONFIG") else None,
    "lsf": None,
    "slurm": None,
}


def config(
    ncpus=None,
    njobs=_config["njobs"],
    configfile=_config["configfile"],
    lsf=_config["lsf"],
    slurm=_config["slurm"],
):
    """
    Function to change HTMD configuration variables.

    Parameters
    ----------
    njobs : int
        Defines the number of parallel jobs spawned for several HTMD operations.
        Negative numbers are used for spawning jobs as many as CPU threads.
        -1: for all CPUs -2: for all except one etc.
    configfile : str
        Defines the HTMD configuration file that is called at the beginning of importing
    lsf : str
        Defines a YAML file that can contain default profile configurations for an LsfQueue
    slurm : str
        Defines a YAML file that can contain default profile configurations for an SlurmQueue
    """
    from jobqueues.config import config as jqconfig

    if ncpus is not None:
        logger.warning(
            "The ncpus config option has been renamed to njobs. Please use njobs instead."
        )
        _config["njobs"] = ncpus
    else:
        _config["njobs"] = njobs
    _config["configfile"] = configfile

    _config["lsf"] = lsf  # DEPRECATE THIS
    _config["slurm"] = slurm  # DEPRECATE THIS
    jqconfig(lsf=lsf, slurm=slurm)



# (c) 2015-2022 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
from glob import glob
from os import path, makedirs
import numpy as np
from htmd.adaptive.adaptive import AdaptiveBase
from htmd.simlist import simlist, simfilter
from htmd.model import Model, macroAccumulate
from protocolinterface import val
from htmd.projections.tica import TICA
from htmd.projections.metric import Metric
import sys
import unittest
import logging

logger = logging.getLogger(__name__)


class AdaptiveMD(AdaptiveBase):
    """Adaptive class which uses a Markov state model for respawning

    AdaptiveMD uses Markov state models to choose respawning poses for the next epochs. In more detail, it projects all
    currently retrieved simulations according to the specified projection, clusters those and then builds a Markov model using
    the discretized trajectories. From the Markov model it then chooses conformations from the various states based on
    the chosen criteria which will be used for starting new simulations.

    Parameters
    ----------
    app : :class:`SimQueue <jobqueues.simqueue.SimQueue>` object, default=None
        A SimQueue class object used to retrieve and submit simulations
    project : str, default='adaptive'
        The name of the project
    nmin : int, default=0
        Minimum number of running simulations
    nmax : int, default=1
        Maximum number of running simulations
    nepochs : int, default=1000
        Stop adaptive once we have reached this number of epochs
    nframes : int, default=0
        Stop adaptive once we have simulated this number of aggregate simulation frames.
    inputpath : str, default='input'
        The directory used to store input folders
    generatorspath : str, default='generators'
        The directory containing the generators
    dryrun : boolean, default=False
        A dry run means that the adaptive will retrieve and generate a new epoch but not submit the simulations
    updateperiod : float, default=0
        When set to a value other than 0, the adaptive will run synchronously every `updateperiod` seconds
    coorname : str, default='input.coor'
        Name of the file containing the starting coordinates for the new simulations
    lock : bool, default=False
        Lock the folder while adaptive is ongoing
    datapath : str, default='data'
        The directory in which the completed simulations are stored
    filter : bool, default=True
        Enable or disable filtering of trajectories.
    filtersel : str, default='not water'
        Atom selection string for filtering.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    filteredpath : str, default='filtered'
        The directory in which the filtered simulations will be stored
    projection : :class:`Projection <moleculekit.projections.projection.Projection>` object, default=None
        A Projection class object or a list of objects which will be used to project the simulation data before constructing a Markov model
    truncation : str, default=None
        Method for truncating the prob distribution (None, 'cumsum', 'statecut'
    statetype : ('micro', 'cluster', 'macro'), str, default='micro'
        What states (cluster, micro, macro) to use for calculations.
    macronum : int, default=8
        The number of macrostates to produce
    skip : int, default=1
        Allows skipping of simulation frames to reduce data. i.e. skip=3 will only keep every third frame
    lag : int, default=1
        The lagtime used to create the Markov model. The units are in frames.
    clustmethod : :class:`ClusterMixin <sklearn.base.ClusterMixin>` class, default=<class 'htmd.clustering.kcenters.KCenter'>
        Clustering algorithm used to cluster the contacts or distances
    method : str, default='1/Mc'
        Criteria used for choosing from which state to respawn from
    ticalag : int, default=20
        Lagtime to use for TICA in frames. When using `skip` remember to change this accordinly.
    ticadim : int, default=3
        Number of TICA dimensions to use. When set to 0 it disables TICA
    contactsym : str, default=None
        Contact symmetry
    save : bool, default=False
        Save the model generated

    Example
    -------
    >>> adapt = AdaptiveMD()
    >>> adapt.nmin = 2
    >>> adapt.nmax = 3
    >>> adapt.nepochs = 2
    >>> adapt.ticadim = 3
    >>> adapt.projection = [MetricDistance('name CA', 'resname MOL', periodic='selections'), MetricDihedral()]
    >>> adapt.generatorspath = htmd.home()+'/data/dhfr'
    >>> adapt.app = LocalGPUQueue()
    >>> adapt.run()
    """

    def __init__(self):
        from sklearn.base import ClusterMixin
        from htmd.clustering.kcenters import KCenter
        from moleculekit.projections.projection import Projection

        super().__init__()
        self._arg(
            "datapath",
            "str",
            "The directory in which the completed simulations are stored",
            "data",
            val.String(),
        )
        self._arg(
            "filter",
            "bool",
            "Enable or disable filtering of trajectories.",
            True,
            val.Boolean(),
        )
        self._arg(
            "filtersel", "str", "Filtering atom selection", "not water", val.String()
        )
        self._arg(
            "filteredpath",
            "str",
            "The directory in which the filtered simulations will be stored",
            "filtered",
            val.String(),
        )
        self._arg(
            "projection",
            ":class:`Projection <moleculekit.projections.projection.Projection>` object",
            "A Projection class object or a list of objects which will be used to project the simulation "
            "data before constructing a Markov model",
            None,
            val.Object(Projection),
            nargs="+",
        )
        self._arg(
            "truncation",
            "str",
            "Method for truncating the prob distribution (None, 'cumsum', 'statecut'",
            None,
            val.String(),
        )
        self._arg(
            "statetype",
            "str",
            "What states (cluster, micro, macro) to use for calculations.",
            "micro",
            val.String(),
            valid_values=("micro", "cluster", "macro"),
        )
        self._arg(
            "macronum",
            "int",
            "The number of macrostates to produce",
            8,
            val.Number(int, "POS"),
        )
        self._arg(
            "skip",
            "int",
            "Allows skipping of simulation frames to reduce data. i.e. skip=3 will only keep every third frame",
            1,
            val.Number(int, "POS"),
        )
        self._arg(
            "lag",
            "int",
            "The lagtime used to create the Markov model",
            1,
            val.Number(int, "POS"),
        )
        self._arg(
            "clustmethod",
            ":class:`ClusterMixin <sklearn.base.ClusterMixin>` class",
            "Clustering algorithm used to cluster the contacts or distances",
            KCenter,
            val.Class(ClusterMixin),
        )
        self._arg(
            "method",
            "str",
            "Criteria used for choosing from which state to respawn from",
            "1/Mc",
            val.String(),
        )
        self._arg(
            "ticalag",
            "int",
            "Lagtime to use for TICA in frames. When using `skip` remember to change this accordinly.",
            20,
            val.Number(int, "0POS"),
        )
        self._arg(
            "ticadim",
            "int",
            "Number of TICA dimensions to use. When set to 0 it disables TICA",
            3,
            val.Number(int, "0POS"),
        )
        self._arg("contactsym", "str", "Contact symmetry", None, val.String())
        self._arg("save", "bool", "Save the model generated", False, val.Boolean())

    def _algorithm(self):
        data = self._getData(self._getSimlist())
        if not self._checkNFrames(data):
            return False
        self._createMSM(data)

        N = self.nmax - self._running
        reward = self._criteria(self._model, self.method)
        reward = self._truncate(reward, N)
        relFrames, _, _ = self._getSpawnFrames(reward, self._model, self._model.data, N)
        self._writeInputs(self._model.data.rel2sim(np.concatenate(relFrames)))
        return True

    def _checkNFrames(self, data):
        if self.nframes != 0 and data.numFrames >= self.nframes:
            logger.info("Reached maximum number of frames. Stopping adaptive.")
            return False
        return True

    def _getSimlist(self):
        logger.info("Postprocessing new data")
        sims = simlist(
            glob(path.join(self.datapath, "*", "")),
            glob(path.join(self.inputpath, "*", "")),
            glob(path.join(self.inputpath, "*", "")),
        )
        if self.filter:
            sims = simfilter(sims, self.filteredpath, filtersel=self.filtersel)
        return sims

    def _on_complete(self):
        if self.filter:
            self._getSimlist()

    def _getData(self, sims):
        metr = Metric(sims, skip=self.skip)
        metr.set(self.projection)

        # if self.contactsym is not None:
        #    contactSymmetry(data, self.contactsym)

        if self.ticadim > 0:
            # tica = TICA(metr, int(max(2, np.ceil(self.ticalag))))  # gianni: without project it was tooooo slow
            data = metr.project()
            data.dropTraj()  # Drop before TICA to avoid broken trajectories
            # 1 < ticalag < (trajLen / 2)
            ticalag = int(
                np.ceil(max(2, min(np.min(data.trajLengths) / 2, self.ticalag)))
            )
            tica = TICA(data, ticalag)
            datadr = tica.project(self.ticadim)
        else:
            datadr = metr.project()
        datadr.dropTraj()  # Preferably we should do this before any projections. Corrupted sims can affect TICA
        return datadr

    def _createMSM(self, data):
        data.cluster(self.clustmethod(n_clusters=self._numClusters(data.numFrames)))
        self._model = Model(data)
        self._model.markovModel(self.lag, self._numMacrostates(data))
        if self.save:
            if not path.exists("saveddata"):
                makedirs("saveddata")
            self._model.save(
                path.join("saveddata", f"e{self._getEpoch()}_adapt_model.dat")
            )

    def _getSpawnFrames(self, reward, model, data, N):
        prob = reward / np.sum(reward)
        logger.debug(f"Sampling probabilities {prob}")
        spawncounts = np.random.multinomial(N, prob)
        logger.debug(f"spawncounts {spawncounts}")

        stateIdx = np.where(spawncounts > 0)[0]
        _, relFrames = model.sampleStates(
            stateIdx, spawncounts[stateIdx], statetype="micro", replacement=True
        )
        logger.debug(f"relFrames {relFrames}")
        return relFrames, spawncounts, prob

    def _criteria(self, model, criteria):
        if criteria == "1/Mc":
            nMicroPerMacro = macroAccumulate(model, np.ones(model.micronum))
            P_I = 1 / macroAccumulate(model, model.data.N[model.cluster_ofmicro])
            P_I = P_I / nMicroPerMacro
            ret = P_I[model.macro_ofmicro]
        elif criteria == "pi/Mc":
            nMicroPerMacro = macroAccumulate(model, np.ones(model.micronum))
            P_I = 1 / macroAccumulate(model, model.data.N[model.cluster_ofmicro])
            P_I = P_I / nMicroPerMacro
            ret = P_I[model.macro_ofmicro] * model.msm.stationary_distribution
        return ret

    def _truncate(self, ranking, N):
        if self.truncation is not None and self.truncation.lower() != "none":
            if self.truncation == "cumsum":
                idx = np.argsort(ranking)
                idx = idx[::-1]  # decreasing sort
                errs = ranking[idx]
                H = (N * errs / np.cumsum(errs)) < 1
                ranking[idx[H]] = 0
            if self.truncation == "statecut":
                idx = np.argsort(ranking)
                idx = idx[::-1]  # decreasing sort
                ranking[idx[N:]] = 0  # Set all states ranked > N to zero.
        return ranking

    def _numClusters(self, numFrames):
        """Heuristic that calculates number of clusters from number of frames"""
        K = int(max(np.round(0.6 * np.log10(numFrames / 1000) * 1000 + 50), 100))
        if K > numFrames / 3:  # Ugly patch for low-data regimes ...
            K = int(numFrames / 3)
        return K

    def _numMacrostates(self, data):
        """Heuristic for calculating the number of macrostates for the Markov model"""
        macronum = self.macronum
        if data.K < macronum:
            macronum = int(np.ceil(data.K / 2))
            logger.warning(
                f"Using less macrostates than requested due to lack of microstates. macronum = {macronum}"
            )

        # Calculating how many timescales are above the lag time to limit number of macrostates
        from deeptime.util.validation import implied_timescales
        from htmd.model import Model

        statelist = [traj.cluster for traj in data.trajectories]
        model = Model._get_model(statelist, self.lag)
        its_data = implied_timescales(model, n_its=macronum)
        timesc = its_data._its
        macronum = min(self.macronum, max(np.sum(timesc > self.lag), 2))
        return macronum


if __name__ == "__main__":
    import htmd.home
    import os
    import shutil
    from htmd.util import tempname
    from moleculekit.projections.metricdistance import MetricDistance

    tmpdir = tempname()
    shutil.copytree(htmd.home.home() + "/data/adaptive/", tmpdir)
    os.chdir(tmpdir)
    md = AdaptiveMD()
    # md.dryrun = True
    md.nmin = 1
    md.nmax = 2
    md.nepochs = 3
    md.ticalag = 2
    md.ticadim = 3
    md.updateperiod = 5
    md.projection = MetricDistance(
        "protein and name CA", "resname BEN and noh", periodic="selections"
    )
    md.projection = [
        MetricDistance(
            "protein and name CA", "resname BEN and noh", periodic="selections"
        ),
        MetricDistance(
            "protein and name CA", "resname BEN and noh", periodic="selections"
        ),
    ]


class _TestAdaptiveRun(unittest.TestCase):
    @unittest.skipIf(sys.platform == "win32", "Windows is not supported")
    def test_adaptive_run(self):
        from moleculekit.projections.metricdistance import MetricSelfDistance
        from jobqueues.simqueue import SimQueue
        from htmd.home import home
        from glob import glob
        import re
        import os
        import tempfile
        import shutil

        with tempfile.TemporaryDirectory() as tmpdir:
            source = os.path.join(tmpdir, "adaptivemd")
            shutil.copytree(os.path.join(home(dataDir="adaptivemd")), source)

            class FakeQueue(SimQueue):
                def __init__(self, source, datadir):
                    super().__init__()
                    self._source = source
                    self._datadir = datadir
                    self._dirs = []

                def retrieve(self):
                    basename = os.path.basename(os.path.abspath(self._dirs[0]))
                    regex = re.compile(r"e(\d+)")
                    epoch = int(regex.search(basename).group(1))
                    for ed, dd in zip(
                        glob(os.path.join(self._source, "data", f"e{epoch}s*")),
                        self._dirs,
                    ):
                        basename = os.path.basename(os.path.abspath(dd))
                        shutil.copytree(ed, os.path.join(self._datadir, basename))

                def submit(self, dirs):
                    self._dirs = dirs

                def inprogress(self):
                    return 0

                def memory(self):
                    return 0

                def ncpu(self):
                    return 1

                def ngpu(self):
                    return 0

                def stop(self):
                    pass

            queue = FakeQueue(source, os.path.join(tmpdir, "data"))

            ad = AdaptiveMD()
            ad.app = queue
            ad.nmin = 1
            ad.nmax = 3
            ad.nepochs = 3
            ad.generatorspath = os.path.join(source, "generators")
            ad.inputpath = os.path.join(tmpdir, "input")
            ad.datapath = os.path.join(tmpdir, "data")
            ad.filteredpath = os.path.join(tmpdir, "filtered")
            protsel = "protein and name CA"
            ad.projection = MetricSelfDistance(protsel)
            ad.updateperiod = 1  # execute every 1 sec
            ad.run()

            assert len(glob(os.path.join(tmpdir, "filtered", "e1*", ""))) == 3
            assert len(glob(os.path.join(tmpdir, "filtered", "e2*", ""))) == 2
            assert len(glob(os.path.join(tmpdir, "filtered", "e3*", ""))) == 2


if __name__ == "__main__":
    unittest.main(verbosity=2)



# (c) 2015-2022 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import re
import abc
import time
import os.path as path
from glob import glob
from os import makedirs
import shutil
from shutil import copytree, ignore_patterns
import numpy as np
from jobqueues.simqueue import RetrieveError, InProgressError, ProjectNotExistError
from htmd.simlist import _simName
from moleculekit.molecule import Molecule
from protocolinterface import ProtocolInterface, val
import logging
import unittest
import os

logger = logging.getLogger(__name__)

_IGNORE_EXTENSIONS = (
    "*.dcd",
    "*.xtc",
    "*.binpos",
    "*.trr",
    "*.nc",
    "*.h5",
    "*.lh5",
    "*.netcdf",
    "*.vel",
    ".done",
    "*.chk",
    "*restart*",
)


class AdaptiveBase(abc.ABC, ProtocolInterface):
    def __init__(self):
        super().__init__()
        from jobqueues.simqueue import SimQueue

        self._arg(
            "app",
            ":class:`SimQueue <jobqueues.simqueue.SimQueue>` object",
            "A SimQueue class object used to retrieve and submit simulations",
            None,
            val.Object((SimQueue,)),
        )
        self._arg("project", "str", "The name of the project", "adaptive", val.String())
        self._arg(
            "nmin",
            "int",
            "Minimum number of running simulations",
            0,
            val.Number(int, "0POS"),
        )
        self._arg(
            "nmax",
            "int",
            "Maximum number of running simulations",
            1,
            val.Number(int, "POS"),
        )
        self._arg(
            "nepochs",
            "int",
            "Stop adaptive once we have reached this number of epochs",
            1000,
            val.Number(int, "POS"),
        )
        self._arg(
            "nframes",
            "int",
            "Stop adaptive once we have simulated this number of aggregate simulation frames.",
            0,
            val.Number(int, "0POS"),
        )
        self._arg(
            "inputpath",
            "str",
            "The directory used to store input folders",
            "input",
            val.String(),
        )
        self._arg(
            "generatorspath",
            "str",
            "The directory containing the generators",
            "generators",
            val.String(),
        )
        self._arg(
            "dryrun",
            "boolean",
            "A dry run means that the adaptive will retrieve and generate a new epoch but not submit the simulations",
            False,
            val.Boolean(),
        )
        self._arg(
            "updateperiod",
            "float",
            "When set to a value other than 0, the adaptive will run synchronously every `updateperiod` seconds",
            0,
            val.Number(float, "0POS"),
        )
        self._arg(
            "coorname",
            "str",
            "Name of the file containing the starting coordinates for the new simulations",
            "input.coor",
            val.String(),
        )
        self._arg(
            "boxname",
            "str",
            "Name of the file containing the starting box dimensions for the new simulations. Set to 'none' to disable box writing.",
            "input.xsc",
            val.String(),
        )
        self._arg(
            "lock",
            "bool",
            "Lock the folder while adaptive is ongoing",
            False,
            val.Boolean(),
        )
        self._arg(
            "mps",
            "int",
            "If mps > 0, it will run simulations using the Multi-Process Service (MPS) with the number of processes specified. If set to 0, mps is disabled",
            0,
            val.Number(int, "0POS"),
        )
        self._running = None

    def run(self):
        """Runs the adaptive

        Use this command to start the adaptive.

        Example
        -------
        >>> adapt = Adaptive()
        >>> adapt.run()
        """
        from natsort import natsorted

        if self.nmax <= self.nmin:
            raise RuntimeError("nmax option should be larger than nmin.")

        self._setLock()
        while True:
            epoch = self._getEpoch()
            logger.info("Processing epoch " + str(epoch))

            if epoch == 0 and self.generatorspath:
                logger.info("Epoch 0, generating first batch")
                self._init()
                input_dirs = natsorted(glob(path.join(self.inputpath, "e1s*")))
                if not self.dryrun:
                    self._submit(input_dirs)
            else:
                # Retrieving simulations
                logger.info("Retrieving simulations.")
                try:
                    self.app.retrieve()
                except RetrieveError as e:
                    logger.error(
                        f"Quitting adaptive run due to error in retrieving simulations: {e}"
                    )
                    return
                except ProjectNotExistError:
                    logger.info(
                        "Retrieve found no previous simulations for this adaptive. Assuming this is a new adaptive run"
                    )

                # Checking how many simulations are in progress (queued/running) on the queue
                try:
                    self._running = self.app.inprogress()
                except InProgressError as e:
                    logger.error(
                        f"Quitting adaptive run due to error in checking number of simulations in progress: {e}"
                    )
                    return
                except ProjectNotExistError:
                    logger.info(
                        "Inprogress found no previous simulations for this adaptive. Assuming this is a new adaptive run"
                    )
                    self._running = 0

                logger.info(str(self._running) + " simulations in progress")

                if epoch >= self.nepochs and self._running == 0:
                    logger.info("Reached maximum number of epochs " + str(self.nepochs))
                    self._on_complete()
                    self._unsetLock()
                    return

                # If currently running simulations are lower than nmin start new ones to reach nmax number of sims
                if self._running <= self.nmin and epoch < self.nepochs:
                    flag = self._algorithm()
                    if flag is False:
                        self._unsetLock()
                        return

                    if not self.dryrun:
                        newsims = glob(
                            path.join(self.inputpath, "e" + str(epoch + 1) + "s*")
                        )
                        try:
                            self._submit(natsorted(newsims))
                        except Exception:
                            # If submitting fails delete all simulation inputs to not confuse _getEpoch()
                            for ns in newsims:
                                shutil.rmtree(ns)
                        logger.info("Finished submitting simulations.")

            if self.updateperiod <= 0:
                break
            logger.info(f"Sleeping for {self.updateperiod} seconds.")
            time.sleep(self.updateperiod)
        self._unsetLock()

    def _setLock(self):
        import datetime

        if self.lock:
            lockfile = os.path.abspath("./adaptivelock")
            if os.path.exists(lockfile):
                raise FileExistsError(
                    "This adaptive folder is locked by a running adaptive application. If this is not"
                    f" the case, delete the {lockfile} file and run adaptive again."
                )

            with open(lockfile, "w") as f:
                f.write("{}".format(datetime.datetime.now()))

    def _unsetLock(self):
        if self.lock:
            lockfile = os.path.abspath("./adaptivelock")
            if os.path.exists(lockfile):
                os.remove(lockfile)

    def _submit(self, newsims):
        from jobqueues.slurmqueue import SlurmQueue

        if self.mps == 0:
            self.app.submit(newsims)
            return

        if not isinstance(self.app, SlurmQueue):
            raise ValueError(
                "Multiple processes per simulation only supported with SlurmQueue."
            )
        for i in range(0, len(newsims), self.mps):
            batch = newsims[i : min(i + self.mps, len(newsims))]
            self.app.submit(batch, nvidia_mps=len(batch) > 1)

    def _init(self):
        from natsort import natsorted

        folders = natsorted(
            glob(path.join(self.generatorspath, "*", ""))
        )  # I need the extra ''  to add a finishing /
        if len(folders) == 0:
            logger.info("Generators folder has no subdirectories, using folder itself")
            folders.append(self.generatorspath)

        numF = len(folders)
        numCopies = np.ones(numF, dtype=int) * int(np.floor(self.nmax / numF))
        numExtra = np.mod(self.nmax, numF)
        extraChoices = np.random.choice(numF, numExtra, replace=False)  # draw the extra
        numCopies[extraChoices] += 1
        # numCopies = numCopies + np.random.multinomial(numExtra, [1/numF]*numF)  # draw the extra equally from a flat distribution
        if not path.exists(self.inputpath):
            makedirs(self.inputpath)

        # Check if epoch 1 directories already exist in the input folder
        existing = glob(path.join(self.inputpath, "e1s*"))
        if len(existing) != 0:
            raise NameError("Epoch 1 directories already exist.")

        k = 1
        for i in range(numF):
            for j in range(numCopies[i]):
                name = _simName(folders[i])
                inputdir = path.join(self.inputpath, "e1s" + str(k) + "_" + name)
                # src = path.join(self.generatorspath, name, '*')
                src = folders[i]
                copytree(
                    src,
                    inputdir,
                    symlinks=True,
                    ignore=ignore_patterns(*_IGNORE_EXTENSIONS),
                )
                k += 1

    def _getEpoch(self):
        """Compute current epoch of adaptive

        Checks the input folder for the latest epoch inputs and returns the epoch number

        Returns
        -------
        epoch : int
            The current epoch
        """
        folders = glob(path.join(self.inputpath, "e*", ""))
        epoch = 0
        regex = re.compile(r"e(\d+)")
        for f in folders:
            res = regex.search(f)
            if res:
                num = res.group(1)
                if int(num) > epoch:
                    epoch = int(num)
        return epoch

    def _writeInputs(self, simsframes, epoch=None):
        if epoch is None:
            epoch = self._getEpoch() + 1

        test = glob(path.join(self.inputpath, "e" + str(epoch) + "*"))
        if len(test) != 0:
            raise NameError("Input dirs of epoch " + str(epoch) + " already exists.")

        from htmd.parallelprogress import ParallelExecutor
        from htmd.config import _config
        from joblib import delayed

        aprun = ParallelExecutor(n_jobs=_config["njobs"])
        aprun(total=len(simsframes), desc="Writing inputs")(
            delayed(_writeInputsFunction)(
                i, f, epoch, self.inputpath, self.coorname, self.boxname
            )
            for i, f in enumerate(simsframes)
        )

    @abc.abstractmethod
    def _algorithm(self):
        return

    def _on_complete(self):
        pass


def _writeInputsFunction(i, f, epoch, inputpath, coorname, boxname):
    regex = re.compile(r"(e\d+s\d+)_")
    frameNum = f.frame
    piece = f.piece
    if f.sim.parent is None:
        currSim = f.sim
    else:
        currSim = f.sim.parent

    traj = currSim.trajectory[piece]
    if currSim.input is None:
        raise NameError(
            "Could not find input folder in simulation lists. Cannot create new simulations."
        )

    wuName = _simName(traj)
    res = regex.search(wuName)
    if (
        res
    ):  # If we are running on top of adaptive, use the first name part for the next sim name
        wuName = res.group(1)

    # create new job directory
    newName = f"e{epoch}s{i+1}_{wuName}p{piece}f{frameNum}"
    newDir = path.join(inputpath, newName, "")

    # copy previous input directory including input files
    copytree(
        currSim.input,
        newDir,
        symlinks=False,
        ignore=ignore_patterns(
            "*.coor", "*.rst", "*.out", "*.xsc", *_IGNORE_EXTENSIONS
        ),
    )

    # overwrite input file with new one. frameNum + 1 as catdcd does 1 based indexing

    # Always read the mol file, otherwise it does not work if we need to save a PDB as coorname
    mol = Molecule(currSim.molfile)
    mol.read(traj)
    mol.dropFrames(keep=frameNum)  # Making sure only specific frame to write is kept
    mol.write(path.join(newDir, coorname))
    if boxname.lower() != "none":
        mol.write(path.join(newDir, boxname))


def epochSimIndexes(simlist):
    """Finds the simulation indexes for each epoch.

    Creates a dictionary with the epoch number as key and values the simlist indexes of the simulations corresponding to
    the given epoch.

    Parameters
    ----------
    simlist : list
        A simulation list created using the :func:`simlist <htmd.simlist.simlist>` function

    Returns
    -------

    """
    epochidx = {}
    for i, sl in enumerate(simlist):
        simepoch = getEpochFromName(sl.trajectory[0])
        if simepoch not in epochidx:
            epochidx[simepoch] = []
        epochidx[simepoch].append(i)
    return epochidx


def getEpochFromName(name):
    """Given a adaptive simulation name, tells you which epoch it belongs to.

    Parameters
    ----------
    name : str
        Simulation name

    Returns
    -------
    epoch : int
        The epoch
    """
    import re

    reg = re.compile(r"/e(\d+)s\d+_")
    matches = reg.findall(name)
    if len(matches) == 0:
        raise RuntimeError(f"{name} is not an adaptive trajectory")
    return int(matches[0])


def reconstructAdaptiveTraj(simlist, trajID):
    """Reconstructs a long trajectory out of short adaptive runs.

    Parameters
    ----------
    simlist : numpy.ndarray of :class:`Sim <htmd.simlist.Sim>` objects
        A simulation list generated by the :func:`simlist <htmd.simlist.simlist>` function
    trajID : int
        The id of the trajectory from which to start going back.

    Returns
    -------
    mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        A Molecule object containing the reconstructed trajectory
    chain : np.ndarray
        The simulation IDs of all simulations involved
    pathlist : np.ndarray of str
        The names of all simulations involved.

    Examples
    --------
    >>> mol, chain, pathlist = reconstructAdaptiveTraj(data.simlist, 52)
    """

    sim = None
    for s in simlist:
        if s.simid == trajID:
            sim = s
            break
    if sim is None:
        raise NameError(f"Could not find sim with ID {trajID} in the simlist.")

    pathlist = []
    pathlist.append(sim.trajectory[0])
    chain = []
    chain.append((sim, -1, -1))

    epo = None
    while epo != 1:
        [sim, piece, frame, epo] = _findprevioustraj(
            simlist, _simName(sim.trajectory[0])
        )
        pathlist.append(sim.trajectory[piece])
        chain.append((sim, piece, frame))
    pathlist = pathlist[::-1]
    chain = chain[::-1]

    mol = Molecule(sim.molfile)
    mol.coords = np.zeros((mol.numAtoms, 3, 0), dtype=np.float32)
    mol.fileloc = []
    mol.box = np.zeros((3, 0), dtype=np.float32)
    for i, c in enumerate(chain):
        tmpmol = Molecule(sim.molfile)
        tmpmol.read(c[0].trajectory)
        endpiece = c[1]
        fileloc = np.vstack(tmpmol.fileloc)
        filenames = fileloc[:, 0]
        pieces = np.unique(filenames)
        firstpieceframe = np.where(filenames == pieces[endpiece])[0][0]
        endFrame = firstpieceframe + c[2]
        if endFrame != -1:
            tmpmol.coords = tmpmol.coords[
                :, :, 0 : endFrame + 1
            ]  # Adding the actual respawned frame (+1) since the respawned sim doesn't include it in the xtc
            tmpmol.fileloc = tmpmol.fileloc[0 : endFrame + 1]
            tmpmol.box = tmpmol.box[:, 0 : endFrame + 1]
        mol.coords = np.concatenate((mol.coords, tmpmol.coords), axis=2)
        mol.box = np.concatenate((mol.box, tmpmol.box), axis=1)
        mol.fileloc += tmpmol.fileloc
    # mol.fileloc[:, 1] = range(np.size(mol.fileloc, 0))

    return mol, chain, pathlist


def _findprevioustraj(simlist, simname):
    regex = re.compile(r"_(e\d+s\d+)p(\d+)f(\d+)$")
    regex2 = re.compile(r"_(e\d+s\d+)f(\d+)$")
    m = regex.search(simname)
    if m:
        prevname = m.group(1)
        prevpiece = int(m.group(2)) - 1
        prevframe = int(m.group(3))
    else:
        m2 = regex2.search(simname)
        if m2:
            prevname = m2.group(1)
            prevpiece = 0
            prevframe = int(m2.group(2))
        else:
            raise NameError(
                f"Could not match simname: {simname} with regular expressions."
            )
    regex = re.compile(r"e(\d+)s")
    m = regex.match(prevname)
    if not m:
        raise NameError(f"Could not parse epoch number from name: {prevname}.")
    epo = int(m.group(1))
    sim = None
    regex = re.compile(f"{prevname}_")
    for s in simlist:
        if len(s.trajectory) <= prevpiece:
            continue
        m = regex.search(s.trajectory[prevpiece])
        if m:
            sim = s
            break
    if sim is None:
        raise NameError(f"Could not find parent of simulation {simname}.")
    return sim, prevpiece, prevframe, epo


class _TestAdaptive(unittest.TestCase):
    def test_input_writer(self):
        from htmd.home import home
        from htmd.simlist import Frame, simlist
        import tempfile

        filedir = home() + "/data/adaptive/"
        sims = simlist(
            glob(os.path.join(filedir, "data", "*", "")),
            glob(os.path.join(filedir, "input", "*", "")),
            glob(os.path.join(filedir, "input", "*", "")),
        )

        with tempfile.TemporaryDirectory() as tmpdir:
            f = Frame(sims[0], 0, 5)
            _writeInputsFunction(1, f, 2, tmpdir, "input.coor", "input.xsc")

            ref = Molecule(sims[0])
            ref.dropFrames(keep=5)
            mol = Molecule(sims[0].molfile)
            mol.read(os.path.join(tmpdir, "e2s2_e1s1p0f5", "input.coor"))
            mol.read(os.path.join(tmpdir, "e2s2_e1s1p0f5", "input.xsc"))

            assert np.array_equal(ref.coords, mol.coords)
            assert np.array_equal(ref.box, mol.box)


if __name__ == "__main__":
    unittest.main(verbosity=2)



# (c) 2015-2022 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
from moleculekit.projections.metricdihedral import MetricDihedral, Dihedral
from htmd.projections.metric import Metric
from moleculekit.molecule import Molecule
from joblib import delayed
import numpy as np
import pandas as pd


class MutualInformation:
    def __init__(self, model, mol=None, fstep=0.1, skip=1):
        """Class that calculates the mutual information of protein residues.

        Parameters
        ----------
        model : :class:`Model <htmd.model.Model>` object
            A Model object with a calculated MSM
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A reference molecule from which to obtain structural information. By default model.data.simlist[0].molfile
            will be used.
        fstep : float
            The frame step of the simulations
        skip : int
            Frame skipping

        Examples
        --------
        >>> from htmd.mutualinformation import MutualInformation
        >>> from htmd.ui import *
        >>>
        >>> sims = simlist(glob('./filtered/*/'), './filtered/filtered.pdb')
        >>> mol = Molecule('./filtered/filtered.pdb')
        >>>
        >>> metr = Metric(sims)
        >>> metr.set(MetricDihedral())
        >>> datadih = metr.project()
        >>> datadih.fstep = 0.1
        >>> datadih.dropTraj()
        >>>
        >>> metr = Metric(datadih.simlist)
        >>> metr.set(MetricSelfDistance('protein and name CA', metric='contacts'))
        >>> dataco = metr.project()
        >>> dataco.fstep = 0.1
        >>> dataco.dropTraj()
        >>>
        >>> tica = TICA(datadih, 20, units='ns')
        >>> datatica = tica.project(3)
        >>> datatica.cluster(MiniBatchKMeans(n_clusters=1500))
        >>> model = Model(datatica)
        >>> model.markovModel(12, 4, units='ns')
        >>>
        >>> mu = MutualInformation(model)
        >>> mu.calculate()
        >>> mu.saveMI('./mi_matrix.npy')
        >>> mu.weightGraph(dataco, 0.005)
        >>> mu.save_graphml('./weightgraph.graphml')
        """
        self.model = model

        self.mol = mol
        if mol is None:
            self.mol = Molecule(self.model.data.simlist[0].molfile)

        self._computeChiDihedrals(fstep=fstep, skip=skip)

        self.bindihcat = self._histogram()  # Lasts two minutes
        self.resids = self.mol.get("resid", "name CA")
        self.residmap = np.ones(self.mol.resid.max() + 1, dtype=int) * -1
        self.residmap[self.resids] = np.arange(len(self.resids))

        self.mi_matrix = None
        self.graph_array = None
        self.graph = None

    def calculate(self, njobs=1):
        """
        Parameters
        ----------
        njobs : int
            Number of parallel jobs to spawn for the calculation of MI
        """
        from htmd.parallelprogress import ParallelExecutor

        numchi = self.chi.numDimensions
        statdist = self.model.msm.stationary_distribution
        stconcat = np.concatenate(self.model.data.St)
        microcat = self.model.micro_ofcluster[stconcat]

        aprun = ParallelExecutor(n_jobs=njobs)
        res = aprun(total=numchi, desc="Calculating MI")(
            delayed(self._parallelAll)(
                numchi, dih1, 4, self.model.micronum, self.bindihcat, microcat, statdist
            )
            for dih1 in range(numchi)
        )
        MI_all = np.zeros((len(self.resids), len(self.resids)))
        for r in res:
            dihcounts = r[0]
            pairs = r[1]
            for dihc, p in zip(dihcounts, pairs):
                dih1, dih2 = p
                if dih1 == dih2:
                    continue
                resid1 = self.residmap[
                    self.mol.resid[self.chi.description.atomIndexes[dih1][0]]
                ]
                resid2 = self.residmap[
                    self.mol.resid[self.chi.description.atomIndexes[dih2][0]]
                ]
                MI_all[resid1][resid2] = self._calcMutualInfo(dihc)
        self.mi_matrix = self._cleanautocorrelations(MI_all)

    def _computeChiDihedrals(self, fstep=0.1, skip=1):
        chis = []
        protmol = self.mol.copy()
        protmol.filter("protein")
        caidx = self.mol.atomselect("protein and name CA")
        resids = self.mol.resid[caidx]
        resnames = self.mol.resname[caidx]
        for residue, resname in zip(resids, resnames):
            ch = Dihedral.chi1(protmol, residue)
            if ch is not None:
                chis.append(ch)

        metr = Metric(self.model.data.simlist, skip=skip)
        metr.set(MetricDihedral(chis, sincos=False))
        data = metr.project()
        data.fstep = fstep
        self.chi = data

    def _histogram(self):
        condata = np.concatenate(self.chi.dat)
        bins = np.array([-180, -150, -90, -30, 0, 30, 90, 150, 180])
        dic = {1: 3, 2: 0, 3: 1, 4: 0, 5: 0, 6: 2, 7: 0, 8: 3, 9: 3}
        binneddih = np.zeros([condata.shape[0], condata.shape[1]])
        for dihedral in range(condata.shape[1]):
            binning = np.digitize(condata[:, dihedral], bins)
            binneddih[:, dihedral] = [dic[n] if n in dic.keys() else n for n in binning]
        return binneddih

    def _calcMutualInfo(self, contingency):
        # Ripped out of sklearn since it converts floats to integers without warning which breaks our use-case
        from math import log

        nzx, nzy = np.nonzero(contingency)
        nz_val = contingency[nzx, nzy]
        contingency_sum = contingency.sum()
        pi = np.ravel(contingency.sum(axis=1))
        pj = np.ravel(contingency.sum(axis=0))
        log_contingency_nm = np.log(nz_val)
        contingency_nm = nz_val / contingency_sum
        # Don't need to calculate the full outer product, just for non-zeroes
        outer = pi.take(nzx) * pj.take(nzy)
        log_outer = -np.log(outer) + log(pi.sum()) + log(pj.sum())
        mi = (
            contingency_nm * (log_contingency_nm - log(contingency_sum))
            + contingency_nm * log_outer
        )
        return mi.sum()

    def _parallelAll(
        self, numdih, dih1, numbins, micronum, bindihcat, microcat, stat_dist
    ):
        results = []
        resultpairs = []
        for dih2 in range(dih1, numdih):
            dihcounts = np.zeros((numbins, numbins, micronum))

            # Find pairs of dihedrals (keys) and which absolute frames they occur in (vals)
            df = pd.DataFrame({"a": list(zip(bindihcat[:, dih1], bindihcat[:, dih2]))})
            gg = df.groupby(by=df.a).groups

            for pair in gg:
                microsofpairs = microcat[
                    gg[pair]
                ]  # Get the microstates of all frames having given dihedral pair
                microsofpairs = np.delete(
                    microsofpairs, np.where(microsofpairs == -1)[0]
                )  # Delete dropped clusters
                counts = np.bincount(
                    microsofpairs
                )  # Count number of frames with that pair for each microstate (0, max_micro_seen)
                dihcounts[
                    int(pair[0]), int(pair[1]), : len(counts)
                ] += counts  # Add the counts
            dihcounts /= dihcounts.sum(axis=0).sum(
                axis=0
            )  # Normalize all slices by total counts in each microstate
            dihcounts *= stat_dist  # Multiply by stationary distribution of each state
            dihcounts = np.sum(
                dihcounts, axis=2
            )  # Calculate the weighted sum over all states
            results.append(dihcounts)
            resultpairs.append((dih1, dih2))

        return results, resultpairs

    def _cleanautocorrelations(self, mi):
        # TODO: vectorize this
        # np.diag(np.ones(3), k=-1).astype(bool) | np.diag(np.ones(3), k=1).astype(bool) | np.diag(np.ones(4)).astype(bool)
        for i in range(mi.shape[0]):
            for j in range(mi.shape[1]):
                if abs(i - j) < 2:
                    mi[i][j] = 0
        return mi

    def saveMI(self, path):
        np.save(path, self.mi_matrix)

    def loadMI(self, path):
        self.mi_matrix = np.load(path)

    def weightGraph(self, datacontacts, mi_threshold, time_treshold=0.6):
        if len(self.mol.get("resid", "name CA")) != len(self.resids):
            raise Exception(
                "The length of the protein doesn't match the Mutual Information data"
            )
        contactcat = np.concatenate(datacontacts.dat)
        contacts_matrix = np.zeros([len(self.resids), len(self.resids)])
        for i in range(contactcat.shape[1]):
            counter = np.count_nonzero(contactcat[:, i])
            resid1 = self.residmap[
                self.mol.resid[datacontacts.description.atomIndexes[i][0]]
            ]
            resid2 = self.residmap[
                self.mol.resid[datacontacts.description.atomIndexes[i][1]]
            ]
            contacts_matrix[resid1][resid2] = counter

        self.graph_array = np.zeros(
            [contacts_matrix.shape[0], contacts_matrix.shape[0]]
        )
        mask = (self.mi_matrix > mi_threshold) & (
            contacts_matrix > (time_treshold * contactcat.shape[0])
        )
        self.graph_array[mask] = self.mi_matrix[mask]

        intermed = []
        for source in range(self.graph_array.shape[0]):
            for target in range(source, self.graph_array.shape[1]):
                if self.graph_array[source, target] != 0 and target > source:
                    intermed.append(
                        [
                            int(self.resids[source]),
                            int(self.resids[target]),
                            float(self.graph_array[source, target]),
                        ]
                    )
        import pandas as pd
        import networkx as nx
        from sklearn.cluster.spectral import SpectralClustering

        pd = pd.DataFrame(intermed, columns=["source", "target", "weight"])
        pd[["source", "target"]] = pd[["source", "target"]].astype(
            type("int", (int,), {})
        )
        pd["weight"] = pd["weight"].astype(type("float", (float,), {}))
        G = nx.from_pandas_edgelist(pd, "source", "target", "weight")
        # setSegment
        segids = self.mol.get("segid", "name CA")
        seg_res_dict = {
            key: value
            for (key, value) in zip(self.resids, segids)
            if np.any(pd.loc[(pd["source"] == key)].index)
            or np.any(pd.loc[(pd["target"] == key)].index)
        }
        nx.set_node_attributes(G, seg_res_dict, "Segment")
        # set
        if not nx.is_connected(G):
            G = max(nx.connected_component_subgraphs(G), key=len)
        flow_cent = nx.current_flow_betweenness_centrality(G, weight="weight")
        nx.set_node_attributes(G, flow_cent, "flowcent")
        Spectre = SpectralClustering(n_clusters=10, affinity="precomputed")
        model = Spectre.fit_predict(self.graph_array)
        model = model.astype(type("float", (float,), {}))
        spectral_dict = {
            key: value for (key, value) in zip(self.resids, model) if key in G.nodes()
        }
        nx.set_node_attributes(G, spectral_dict, "spectral")
        self.graph = G

    def save_graphml(self, path):
        import networkx as nx

        nx.write_graphml(self.graph, path)

    # def save_pdf(self, graphpath, outpath):
    #     print(self.graph_array)
    #     self.save_graphml(graphpath)
    #     compile_java('./GephiGraph.java')
    #     compile_java('./GraphTest.java')
    #     execute_java('GraphTest', graphpath, outpath)


# if __name__ == '__main__':
#     from htmd.mutualinformation import MutualInformation
#     from htmd import *
#
#     sims = simlist(glob('/shared/adria/2ov5/adaptive_amber/batches/1/filtered/*/'),
#                    '/shared/adria/2ov5/adaptive_amber/batches/1/filtered/filtered.pdb')
#     mol = Molecule('/shared/adria/2ov5/adaptive_amber/batches/1/filtered/filtered.pdb')
#
#     metr = Metric(sims[0:100])
#     metr.set(MetricDihedral())
#     datadih = metr.project()
#     datadih.fstep = 0.1
#     datadih.dropTraj()
#
#     metr = Metric(datadih.simlist)
#     metr.set(MetricSelfDistance('protein and name CA', metric='contacts', threshold=8))
#     dataco = metr.project()
#     dataco.fstep = 0.1
#     dataco.dropTraj()
#
#     tica = TICA(datadih, 20, units='ns')
#     datatica = tica.project(3)
#     datatica.cluster(MiniBatchKMeans(n_clusters=1500))
#     model = Model(datatica)
#     model.markovModel(12, 4, units='ns')
#
#     mu = MutualInformation(model)
#     mu.calculate()
#     mu.saveMI('/tmp/mi_matrix.npy')
#     mu.weightGraph(dataco, 0.005)
#     mu.save_graphml('/tmp/weightgraph_0-005.graphml')



# (c) 2015-2022 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import numpy as np
from scipy.spatial.distance import cdist
from sklearn.base import BaseEstimator, ClusterMixin, TransformerMixin
import logging

logger = logging.getLogger(__name__)


class KCenter(BaseEstimator, ClusterMixin, TransformerMixin):
    """Class to perform KCenter clustering of a given data set

    KCenter randomly picks one point from the data, which is now the center of the first cluster. All points are put
    into the first cluster. In general the furthest point from its center is chosen to be the new center. All points,
    which are closer to the new center than the old one are assigned to the new cluster. This goes on, until K clusters
    have been created.

    Parameters
    ----------
    n_clusters: int
        desired number of clusters

    Examples
    --------
    >>> cluster = KCenter(n_cluster=200)
    >>> cluster.fit(data)

    Attributes
    ----------
    cluster_centers : list
        list with the points, which are the centers of the clusters
    centerFrames : list
        list of indices of center points in data array
    labels_ : list
        list with number of cluster of each frame
    clusterSize_ : list
        list with number of frames in each cluster
    distance : list
        list with the distance of each frame from the nearest center
    """

    def __init__(self, n_clusters):
        self.n_clusters = n_clusters
        self.cluster_centers_ = []
        self.centerFrames = []
        self.labels_ = []
        self.clusterSize = []
        self.distance = []

    def fit(self, data):
        """Compute the centroids of data.

        Parameters
        ----------
        data : np.ndarray
            A 2D array of data. Columns are features and rows are data examples.
        """
        if len(self.cluster_centers_) != 0:
            logger.warning("Clustering already exists. Reclustering data!")
            self.cluster_centers_ = []
            self.centerFrames = []
            self.clusterSize = []

        # Initialization
        # select random point and assign all points to cluster 0
        numpoints = np.size(data, 0)

        idxCenter = np.random.randint(numpoints)
        self.cluster_centers_.append(data[idxCenter, :])
        self.centerFrames.append(idxCenter)
        self.labels_ = np.zeros(numpoints, dtype=int)

        dist = self._dist(self.cluster_centers_, data)
        countCluster = 1

        while len(self.cluster_centers_) < self.n_clusters:
            if np.max(dist) == 0:
                break

            # find point furthest away from all center
            newCenterIdx = np.argmax(dist)
            newCenter = data[newCenterIdx, :]
            self.centerFrames.append(newCenterIdx)
            self.cluster_centers_.append(newCenter)

            # find all points closer to new center than old center
            newdist = self._dist(newCenter, data)
            switchIdx = dist > newdist

            # assign them to new cluster
            self.labels_[switchIdx] = countCluster
            dist[switchIdx] = newdist[switchIdx]

            countCluster += 1

        # update clusterSize
        self.clusterSize = np.bincount(self.labels_)
        self.distance = dist
        self.cluster_centers_ = np.array(self.cluster_centers_)

    @staticmethod
    def _dist(centers, data):
        dist = np.squeeze(cdist(np.atleast_2d(data), np.atleast_2d(centers)))
        return dist


if __name__ == "__main__":
    """
    infile = open("../../clusterdata/R15.txt")
    data = []
    xdata = []
    ydata =[]
    properLabels = []

    for line in infile:
        stuff = line.split()
        data.append((float(stuff[0]), float(stuff[1])))
        xdata.append(float(stuff[0]))
        ydata.append(float(stuff[1]))
        properLabels.append(stuff[2])

    K = len(set(properLabels))
    """

    data = np.random.rand(100, 10)

    cluster1 = KCenter(n_clusters=20)
    cluster1.fit(data)

    """
    from matplotlib import pylab as plt
    plt.figure(0)
    plt.scatter(xdata, ydata, c=cluster1.labels_)

    xcenters1 = []
    ycenters1 = []
    xcenters2 = []
    ycenters2 = []
    for center in cluster1.cluster_centers_:
        xcenters1.append(center[0])
        ycenters1.append(center[1])

    plt.figure(0)
    plt.title("KCenter K="+str(K))
    plt.scatter(xcenters1, ycenters1, marker='H', s=100, c='red')
    plt.show()
    """



# (c) 2015-2021 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import json
import os
import platform
import requests
from subprocess import call
import sys


def htmd_show_news():
    try:
        res = requests.get("https://www.htmd.org/news/content", timeout=3)
        print(res.text)
    except Exception:
        pass


def _get_reg_file(product):

    dir = os.path.join(os.path.expanduser("~"), "." + product, ".registered-" + product)
    os.makedirs(dir, exist_ok=True)
    file = os.path.join(dir, "registration")

    return file


def _check_registration(product):

    # Get a registation file
    reg_file = _get_reg_file(product)

    if not os.path.exists(reg_file):
        print("Registation file does not exist")
        return False

    data = {"product": product}
    try:
        with open(reg_file) as fh:
            data.update(json.load(fh))
    except Exception:
        print("Cannot read the registration file")
        return False

    # Note: remove after switching to the new backend
    if "code" not in data and "ret" in data:
        data["code"] = data["ret"]

    # Send the registration data
    try:
        url = "https://registration.acellera.com/check"
        res = requests.post(url, data, timeout=10)
    except Exception as e:
        print(
            f"Could not contact registration server: {e}. Registration is not approved."
        )
        return False

    # Check the response
    if res.status_code == 200:
        status = res.json()
        if "approved" in status:
            return True

        # Note: remove after switching to the new backend
        if "pending" in status:
            return True

    print(f"Registration is not approved: {res.text}")
    return False


def _accept_licence(product):

    ret = input("Type 'yes' to accept the license [no] : ").strip()
    if ret != "yes":
        sys.exit(1)


def _show_licence(product):

    # Find the licence file
    path = os.path.join(os.path.dirname(__file__), "..", product, "LICENCE.txt")
    if not os.path.exists(path):
        print("Cannot find the licence file!")
        sys.exit(1)

    # Show the licence file
    # if sys.stdout.isatty() and sys.stdin:
    #     if platform.system() == "Windows":
    #         call(["c:\\Windows\\System32\\more.com", path])
    #     else:
    #         call(["more", path])
    # else:
    with open(path) as fh:
        print(fh.read())


def check_for_license():
    from pathlib import Path

    envvars = [
        "ACELLERA_LICENCE_SERVER",
        "ACELLERA_LICENSE_SERVER",
        "ACELLERA_LICENCE_FILE",
        "ACELLERA_LICENSE_FILE",
    ]
    user_home = str(Path.home())
    locations = [
        "/opt/acellera/licence.dat",
        "/opt/acellera/license.dat",
        os.path.join(user_home, ".acellera/licence.dat"),
        os.path.join(user_home, ".acellera/license.dat"),
    ]

    if any([envv in os.environ for envv in envvars]) or any(
        [os.path.exists(ll) for ll in locations]
    ):
        from htmd.home import home
        import subprocess

        try:
            ret = subprocess.Popen(
                os.path.join(home(shareDir=True), "license-checker"),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            ret.wait()
            if ret.returncode == 0:
                return True
            else:  # Only print if license check failed
                print(ret.communicate()[0].decode("utf-8"))
        except Exception:
            pass
    return False


def htmd_registration(product="htmd"):
    if not _check_registration(product):
        if check_for_license():
            return
        _show_licence(product)
        print(
            "By continuing to use HTMD you are automatically accepting the above license agreement.\n"
            "For commercial licenses please contact info@acellera.com\n"
            "To remove the above license message please register HTMD by calling htmd_register()"
        )


def _ask(prompt):

    value = ""
    while value == "":
        value = input(prompt).strip()

    return value


def htmd_register(product="htmd"):

    print(
        """
  Welcome to the HTMD registration!

  We would like to know about you to keep in touch.
  Please provide your full name, institutional email,
  institution name, city, and country.
"""
    )

    # Ask a user for data
    data = {}
    data["name"] = _ask("  Full name           : ")
    data["email"] = _ask("  Institutional email : ")
    data["institution"] = _ask("  Institution name    : ")
    data["city"] = _ask("  City                : ")
    data["country"] = _ask("  Country             : ")
    data["product"] = product

    # Send data to the registration server
    url = "https://registration.acellera.com/register"
    res = requests.post(url, data=data, timeout=10)

    # Check the response
    if res.status_code == 200:
        reg_file = _get_reg_file(product)
        with open(reg_file, "w") as fh:
            fh.write(res.text)
        print("\n  Registration completed!\n")
    else:
        print(f"\n  Registration failed: {res.text}\n")


if __name__ == "__main__":
    pass



# (c) 2015-2022 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import numpy as np
from scipy.spatial import cKDTree
from moleculekit.molecule import Molecule
from moleculekit.util import sequenceID
import logging

logger = logging.getLogger(__name__)

# Info about all ions, with charge, name, amber and charmm names (may
# need residues too). Only the first column is currently used.
_ions = {
    "NA": (1, "sodium", "Na+", "SOD"),
    "MG": (2, "magnesium", "Mg2+", "MG"),
    "ZN": (2, "zinc", "Zn2+", "ZN"),
    "K": (1, "potassium", "K+", "POT"),
    "CS": (1, "cesium", "Cs+", "CES"),
    "CA": (2, "calcium", "Ca2+", "CAL"),
    "CL": (-1, "chloride", "Cl-", "CLA"),
}

_ion_aliases = {}
for _key, (_charge, _fullname, _ambername, _charmmname) in _ions.items():
    for _alias in (_key, _fullname, _ambername, _charmmname):
        _ion_aliases[_alias.upper()] = _key


def ionize(
    mol,
    netcharge,
    nwater,
    neutralize=True,
    saltconc=None,
    cation=None,
    anion=None,
    ff=None,
):
    if ff:
        logger.warning(
            "The ff option is no longer needed and is deprecated. "
            "Please use standard names (e.g. NA, CL, K) for `cation` and `anion` arguments."
        )

    if cation is None:
        cation = "NA"
    else:
        cation = _resolve_ion(cation)
    if anion is None:
        anion = "CL"
    else:
        anion = _resolve_ion(anion)

    if saltconc is not None:
        neutralize = False

    cationcharge = _ionGetCharge(cation)
    anioncharge = _ionGetCharge(anion)

    roundnetcharge = int(np.round(netcharge))
    done = False
    ncation = 0
    nanion = 0

    # First we neutralize the system
    if netcharge > 0:
        if anioncharge == -1:
            nanion = roundnetcharge
            ncation = 0
        elif anioncharge == -2:
            if np.mod(roundnetcharge, 2) == 0:  # Total charge is even
                nanion = int(roundnetcharge / 2)
                ncation = 0
            else:  # The charge is odd
                if cationcharge == 1:
                    nanion = 1 + int(np.floor(roundnetcharge / 2))
                    ncation = 1
                else:
                    raise NameError(
                        "Failed to neutralize because the cation and anion charges cannot reach 0"
                    )
    elif netcharge < 0:
        roundnetchargeabs = abs(roundnetcharge)
        if cationcharge == 1:
            nanion = 0
            ncation = roundnetchargeabs
        elif cationcharge == 2:
            if np.mod(roundnetchargeabs, 2) == 0:
                nanion = 0
                ncation = int(roundnetchargeabs / 2)
            else:
                if anioncharge == -1:
                    nanion = 1
                    ncation = 1 + int(np.floor(roundnetchargeabs / 2))
                else:
                    raise NameError(
                        "Failed to neutralize because the cation and anion charges cannot reach 0"
                    )
    else:
        logger.info("The system is already neutral.")
        done = True

    if not done:
        newtotalcharge = nanion * anioncharge + ncation * cationcharge + roundnetcharge
        if newtotalcharge != 0:
            raise NameError("Could not neutralize the system")
    if neutralize:
        return anion, cation, nanion, ncation

    # If we want a specific salt concentration, guess chemical formula
    if cationcharge == 1 and anioncharge == -1:  # e.g., NaCl, Kcl, ...
        cationstoich = 1
        anionstoich = 1
    elif cationcharge == 2 and anioncharge == -1:  # e.g., MgCl2
        cationstoich = 1
        anionstoich = 2
    elif cationcharge == 1 and anioncharge == -2:
        cationstoich = 2
        anionstoich = 1
    elif cationcharge == 2 and anioncharge == -2:
        cationstoich = 1
        anionstoich = 1
    else:
        raise NameError("Unsupported ion charge; cannot guess chamical formula.")

    # Count existing salt concentration to not add on top of existing conc
    # TODO: This will break on multi-atom ions! Will require a bit fancier code to calculate individual molecules
    cation_names = [cation, _ions[cation][3]]  # AMBER and CHARMM names
    anion_names = [anion, _ions[anion][3]]
    ncations_exist = np.sum(np.isin(mol.resname, cation_names)) / cationstoich
    nanions_exist = np.sum(np.isin(mol.resname, anion_names)) / anionstoich
    exist = int(min(ncations_exist, nanions_exist))

    num = int(np.floor(0.5 + 0.0187 * saltconc * nwater))
    num -= exist
    if num < 0:
        raise RuntimeError(
            f"Salt concentration in system is already higher than the requested concentration {saltconc} M. Please remove some ions or change salt concentration and try again."
        )

    logger.info(
        f"Adding {nanion} anions + {ncation} cations for neutralizing and {(cationstoich + anionstoich) * num} ions for the given salt concentration {saltconc} M."
    )
    ncation += cationstoich * num
    nanion += anionstoich * num
    return anion, cation, anion, cation, nanion, ncation


def _resolve_ion(name):
    """Resolve any ion alias (canonical key, full name, AMBER name, CHARMM name) to the canonical key."""
    key = _ion_aliases.get(name.upper())
    if key is None:
        valid = ", ".join(
            f"{k} ({v[1]}, {v[2]}, {v[3]})" for k, v in _ions.items()
        )
        raise NameError(
            f"Ion '{name}' not recognized. Valid ions and their aliases: {valid}"
        )
    return key


def _ionGetCharge(ion):
    ion = _resolve_ion(ion)
    return _ions[ion][0]


def ionizePlace(
    solvent_mol,
    solute_mol,
    anion_resname,
    cation_resname,
    anion_name,
    cation_name,
    nanion,
    ncation,
    dfrom=5,
    dbetween=5,
    segname=None,
):
    """Place a given number of negative and positive ions in the solvent.

    Uses farthest point sampling to produce well-spaced ion positions and
    randomly assigns ion types for uniform spatial mixing.
    Replaces water molecules as long as they respect the given distance criteria.

    Parameters
    ----------
    solvent_mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        The solvent (water) molecule
    solute_mol : :class:`Molecule <moleculekit.molecule.Molecule>` object or None
        The solute molecule used for minimum-distance filtering. If None or
        empty, no minimum distance from solute is enforced.
    anion_resname : str
        Resname of the added anions
    cation_resname : str
        Resname of the added cations
    anion_name : str
        Name of the added anions
    cation_name : str
        Name of the added cations
    nanion : int
        Number of anions to add
    ncation : int
        Number of cations to add
    dfrom : float
        Min distance of ions from solute molecule
    dbetween : float
        Min distance between ions
    segname : str
        Segment name to add

    Returns
    -------
    mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        The solvent molecule with ions replacing some water molecules
    """

    newmol = solvent_mol.copy()

    logger.debug(f"Min distance of ions from molecule: {dfrom}A")
    logger.debug(f"Min distance between ions: {dbetween}A")
    logger.debug(f"Placing {nanion} anions and {ncation} cations.")

    if (nanion + ncation) == 0:
        return newmol

    nions = nanion + ncation

    betabackup = newmol.beta.copy()
    newmol.set("beta", sequenceID((newmol.resid, newmol.insertion, newmol.segid)))

    # Find water oxygens to replace with ions
    wat_oh = newmol.atomselect("noh and water", indexes=True)

    if len(wat_oh) == 0:
        raise NameError("No water molecules found in the solvent to replace with ions.")

    wat_coords = newmol.coords[wat_oh, :, 0]

    # Filter by minimum distance from solute using KD-tree
    if solute_mol is not None and solute_mol.numAtoms > 0:
        solute_tree = cKDTree(solute_mol.coords[:, :, 0])
        dists_to_solute, _ = solute_tree.query(wat_coords)
        valid_mask = dists_to_solute >= dfrom
        valid_local = np.where(valid_mask)[0]
    else:
        valid_local = np.arange(len(wat_oh))

    if len(valid_local) == 0:
        raise NameError(
            f"No waters could be found further than {dfrom} from other molecules to be replaced by ions. You might need to "
            "solvate with a bigger box or disable the ionize property when building."
        )

    valid_coords = wat_coords[valid_local]

    # Farthest point sampling: iteratively pick the water farthest from all
    # previously placed ions.  This is deterministic, requires no retries,
    # and naturally maximises the minimum inter-ion distance.
    min_dists = np.full(len(valid_coords), np.inf)
    selected = []
    used = np.zeros(len(valid_coords), dtype=bool)

    for i in range(nions):
        idx = np.argmax(min_dists)
        if i > 0 and min_dists[idx] < dbetween:
            raise NameError(
                f"Could not place all ions while maintaining minimum distance of {dbetween} A between ions. Try decreasing the between parameter, "
                "decreasing ion concentration or making a larger water box."
            )
        selected.append(idx)
        used[idx] = True

        # Update min distances from the newly placed ion
        dists_to_new = np.linalg.norm(valid_coords - valid_coords[idx], axis=1)
        min_dists = np.minimum(min_dists, dists_to_new)
        min_dists[used] = -np.inf

    # Map selected indices back to atom indices in newmol
    ionlist = [wat_oh[valid_local[s]] for s in selected]

    # Randomly assign ion types to FPS positions.  Since FPS guarantees
    # well-spaced positions regardless of type, random labelling gives
    # uniform spatial mixing independent of the anion/cation ratio.
    type_labels = np.array(["anion"] * nanion + ["cation"] * ncation)
    np.random.shuffle(type_labels)

    # Delete waters but keep their coordinates
    waterpos = np.atleast_2d(newmol.get("coords", ionlist))
    betasel = np.zeros(newmol.numAtoms, dtype=bool)
    for b in newmol.beta[ionlist]:
        betasel |= newmol.beta == b

    sel = np.where(betasel)[0]
    newmol.remove(sel, _logger=False)
    betabackup = np.delete(betabackup, sel)

    # Build all ions as a single Molecule and append once
    ions_mol = Molecule()
    ions_mol.empty(nions)
    ions_mol.chain[:] = "I"
    ions_mol.segid[:] = "I"
    ions_mol.coords = waterpos[:, :, np.newaxis]

    anion_mask = type_labels == "anion"
    ions_mol.name[anion_mask] = anion_name
    ions_mol.resname[anion_mask] = anion_resname
    ions_mol.name[~anion_mask] = cation_name
    ions_mol.resname[~anion_mask] = cation_resname

    base_resid = newmol.resid[-1] + 1
    ions_mol.resid[:] = np.arange(base_resid, base_resid + nions)

    newmol.append(ions_mol)

    # Restoring the original betas
    newmol.beta[: len(betabackup)] = betabackup
    return newmol


def _getSegname(mol, segname):
    defsegnames = [
        "ION",
        "ION1",
        "ION3",
        "ION4",
        "ION5",
        "ION6",
        "ION7",
        "ION8",
        "ION9",
        "IN1",
        "IN2",
        "IN3",
        "IN4",
        "IN5",
        "IN6",
        "IN7",
        "IN8",
        "IN9",
    ]

    # Make sure segname is not taken
    if segname is None:
        for i in range(len(defsegnames)):
            sel = mol.segid == defsegnames[i]
            if not np.any(sel):
                segname = defsegnames[i]
                break
        if segname is None:
            raise NameError("All default segnames taken! Provide your own.")
    else:
        sel = mol.segid == segname
        if np.any(sel):
            raise NameError("Segname already taken. Provide a different one.")



# (c) 2015-2022 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
from htmd.util import tempname
from subprocess import call
import numpy as np
from moleculekit.molecule import Molecule
import shutil


def loopModeller(
    mol, segid, seq, startresid, movstart=None, movend=None, modellerexe="mod9.18"
):
    """Uses the Modeller software to predict missing loops in a Molecule.

    Parameters
    ----------
    mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        A Molecule object.
    segid : str
        The name of the segment containing the gap.
    seq : str
        The sequence of residues to be added by the loop modeller.
    startresid : int
        The resid of the residue before the gap.
    movstart : int
        The resid after which the residues will be allowed to move in modeller.
    movend : int
        The resid before which the residues will be allowed to move in modeller.
    modellerexe : str
        The path to the modeller executable.

    Returns
    -------
    newmol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        A new Molecule object containing the protein with the modelled loop.

    Examples
    --------
    >>> mol = Molecule('1qg8')
    >>> mol2 = loopModeller(mol, '0', 'ENR', 133)
    """
    if shutil.which(modellerexe) is None:
        raise NameError(
            "Could not find a Modeller executable called `{}` in the PATH. This might indicate a wrong path"
            " to the executable or a missing installation. To install modeller use `conda install -c "
            "salilab modeller` and follow the instructions. To provide the correct path change the "
            "`modellerexe` argument".format(modellerexe)
        )

    if movstart is None:
        movstart = startresid
    if movend is None:
        movend = movstart + len(seq) + 1

    segatm = mol.segid == segid
    segres = np.unique(mol.resid[segatm])
    chain = np.unique(mol.chain[segatm])
    if len(chain) != 1:
        raise RuntimeError("More than one chain detected in segment {}".format(segid))
    pos = np.where(segres == startresid)[0][0]
    if (segres[pos + 1] - segres[pos]) != (len(seq) + 1):
        raise RuntimeError(
            "Sequence is {} characters long while sequence gap ({}-{}) is {} long. Cannot insert loop.".format(
                len(seq), segres[pos], segres[pos + 1], (segres[pos + 1] - segres[pos])
            )
        )

    chain = chain[0]
    segresidstart = np.min(segres)
    segresidend = np.max(segres)

    currseq = mol.sequence()[segid]
    minuses = "".join(["-"] * len(seq))
    inspos = np.where(segres == startresid)[0][0] + 1
    gapseq = currseq[:inspos] + minuses + currseq[inspos:] + "*"
    fullseq = currseq[:inspos] + seq + currseq[inspos:] + "*"

    # Get the sequence of the 1qg8 PDB file, and write to an alignment file
    pdbfile = tempname(suffix=".pdb")
    mol.write(pdbfile)

    alifile = tempname(suffix=".ali")
    f = open(alifile, "w")
    f.write(
        ">P1;prot\nstructure:{}:{}:{}:{}:{}::::\n".format(
            pdbfile, segresidstart, chain, segresidend, chain
        )
    )
    f.write(gapseq)
    f.write("\n>P1;prot_fill\nsequence:::::::::\n")
    f.write(fullseq)
    f.close()

    script = """
from modeller import *
from modeller.automodel import *    # Load the automodel class

log.verbose()
env = environ()

class MyModel(automodel):
    def select_atoms(self):
        return selection(self.residue_range('{}', '{}'))

a = MyModel(env, alnfile = '{}', knowns = 'prot', sequence = 'prot_fill')
a.starting_model= 1
a.ending_model  = 1

a.make()
    """.format(
        movstart + 1 - segresidstart, movend - segresidstart, alifile
    )
    pyfile = tempname(suffix=".py")
    f = open(pyfile, "w")
    f.write(script)
    f.close()

    print(pdbfile)
    print(alifile)
    print(pyfile)

    call("{} {}".format(modellerexe, pyfile), shell=True)
    newmol = Molecule("./prot_fill.B99990001.pdb")
    print(
        "You can ignore the `import site` error (https://salilab.org/modeller/release.html#issues)."
        " The results should have been returned."
    )
    return newmol


"""
def loopmodelerFALC(mol, segid, seq, pos, inspos, outname='loop'):
    currseq = mol.sequence()[segid]
    fullseq = currseq[:inspos] + seq + currseq[inspos:]

    seqfile = tempname(suffix='.fasta')
    f = open(seqfile, 'w')
    f.write(fullseq)
    f.close()

    ulrfile = tempname(suffix='.ulr')  # Unreliable local region file: http://galaxy.seoklab.org/softwares/falc.html
    f = open(ulrfile, 'w')
    f.write('1 L {}-{} 0.000'.format(pos, pos+len(seq)+1))
    f.close()

    pdbfile = tempname(suffix='.pdb')
    mol.write(pdbfile)

    call('/shared/sdoerr/Software/FALC/bin/falc.py -title {} -pdb {} -fa {} -ulr {} -n {}'.format(outname, pdbfile, seqfile, ulrfile, 1))
"""

if __name__ == "__main__":
    pass
    # from moleculekit.molecule import Molecule
    # mol = Molecule('1qg8')
    # mol2 = loopModeller(mol, '0', 'ENR', 133)



# (c) 2015-2022 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
try:
    import openmm
except ImportError:
    raise ImportError(
        "openmm is not installed. Please install it using `pip install openmm`."
    )
from openmm import unit
from openmm import app
import numpy as np


def _halton_sequence(p, n):
    """
    Halton deterministic sequence on [0,1].
    Parameters
    ----------
    p : int
       Prime number for sequence.
    n : int
       Sequence length to generate.
    Returns
    -------
    u : numpy.array of double
       Sequence on [0,1].
    Notes
    -----
    Code source: http://blue.math.buffalo.edu/sauer2py/
    More info: http://en.wikipedia.org/wiki/Halton_sequence
    Examples
    --------
    Generate some sequences with different prime number bases.
    >>> x = _halton_sequence(2,100)
    >>> y = _halton_sequence(3,100)
    >>> z = _halton_sequence(5,100)
    """
    eps = np.finfo(np.double).eps
    # largest number of digits (adding one for halton_sequence(2,64) corner case)
    b = np.zeros(int(np.ceil(np.log(n) / np.log(p))) + 1)
    u = np.empty(n)
    for j in range(n):
        i = 0
        b[0] += 1  # add one to current integer
        while b[i] > p - 1 + eps:  # this loop does carrying in base p
            b[i] = 0
            i = i + 1
            b[i] += 1
        u[j] = 0
        for k in range(len(b)):  # add up reversed digits
            u[j] += b[k] * p ** -(k + 1)
    return u


def _subrandom_particle_positions(nparticles, box_vectors, ndim, method="halton"):
    """Generate a deterministic list of subrandom particle positions."""
    # Create positions array.
    positions = np.zeros([nparticles, 3], np.float32)

    # Fill in each dimension.
    primes = [2, 3, 5]  # prime bases for Halton sequence
    for dim in range(ndim):
        x = _halton_sequence(primes[dim], nparticles)
        ll = box_vectors[dim][dim]
        positions[:, dim] = x - 0.5

    np.random.shuffle(positions)
    return unit.Quantity(positions * ll / ll.unit, ll.unit)


def distributeLipids(
    boxsize,
    resnames,
    sigmas,
    cutoff,
    mass=39.9 * unit.amu,  # argon
    epsilon=0.238 * unit.kilocalories_per_mole,  # argon,
    switch_width=3.4 * unit.angstrom,  # argon
):
    from moleculekit.periodictable import periodictable

    nparticles = len(resnames)

    # Determine Lennard-Jones cutoff.
    cutoff = cutoff * unit.angstrom

    cutoff_type = openmm.NonbondedForce.CutoffPeriodic

    # Create an empty system object.
    system = openmm.System()

    # Periodic box vectors.
    a = unit.Quantity(
        (boxsize[0] * unit.angstrom, 0 * unit.angstrom, 0 * unit.angstrom)
    )
    b = unit.Quantity(
        (0 * unit.angstrom, boxsize[1] * unit.angstrom, 0 * unit.angstrom)
    )
    c = unit.Quantity(
        (0 * unit.angstrom, 0 * unit.angstrom, boxsize[2] * unit.angstrom)
    )
    system.setDefaultPeriodicBoxVectors(a, b, c)

    # Set up periodic nonbonded interactions with a cutoff.
    nb = openmm.NonbondedForce()
    nb.setNonbondedMethod(cutoff_type)
    nb.setCutoffDistance(cutoff)
    nb.setUseDispersionCorrection(True)

    nb.setUseSwitchingFunction(False)
    if switch_width is not None:
        nb.setUseSwitchingFunction(True)
        nb.setSwitchingDistance(cutoff - switch_width)

    for s in sigmas:
        system.addParticle(mass)
        nb.addParticle(0.0 * unit.elementary_charge, s * unit.angstrom, epsilon)

    positions = _subrandom_particle_positions(
        nparticles, system.getDefaultPeriodicBoxVectors(), 2
    )

    # Add the nonbonded force.
    system.addForce(nb)

    # Add a restraining potential to keep atoms in z=0
    energy_expression = "k * (z^2)"
    force = openmm.CustomExternalForce(energy_expression)
    force.addGlobalParameter("k", 10)
    for particle_index in range(nparticles):
        force.addParticle(particle_index, [])
    system.addForce(force)

    # Create topology.
    topology = app.Topology()
    chain = topology.addChain()
    elems = list(periodictable.keys())  # Just create residues of random dummy elements
    _, idx = np.unique(resnames, return_inverse=True)
    for i in idx:
        element = app.Element.getBySymbol(elems[i])
        residue = topology.addResidue(elems[i], chain)
        topology.addAtom(elems[i], element, residue)

    topology.setUnitCellDimensions(unit.Quantity(boxsize, unit.angstrom))

    # Simulate it
    from openmm import VerletIntegrator
    from openmm.app import Simulation
    from openmm.unit import picoseconds, angstrom

    nsteps = 10000

    integrator = VerletIntegrator(0.002 * picoseconds)
    simulation = Simulation(topology, system, integrator)
    simulation.context.setPositions(positions)
    simulation.minimizeEnergy()
    # simulation.reporters.append(DCDReporter('output.dcd', 1))
    # simulation.reporters.append(StateDataReporter(stdout, 1000, potentialEnergy=True, totalEnergy=True, step=True, separator='   '))
    simulation.step(nsteps)

    state = simulation.context.getState(getPositions=True, enforcePeriodicBox=True)
    allfinalpos = state.getPositions(asNumpy=True).value_in_unit(angstrom)

    return allfinalpos


if __name__ == "__main__":
    pass



"""
Markov state models are a statistical tool for analysing molecular simulations which has met with lots of success.
The Model class here, encapsulates all functionallity for the calculation of Markov models while hiding unnecessary
details under the hood. It uses DeepTime [1]_ internally to calculate Markov models.

References
----------
.. [1] Deeptime: a Python library for machine learning dynamical models from time series data. Moritz Hoffmann et al 2022 Mach. Learn.: Sci. Technol. 3 015009
"""

# (c) 2015-2022 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import numpy as np
import random
from moleculekit.molecule import Molecule
from htmd.units import convert as unitconvert
import unittest
import logging

logger = logging.getLogger(__name__)


class Model(object):
    """Constructor for the Model class.

    Parameters
    ----------
    data : :class:`MetricData <htmd.metricdata.MetricData>` object
        A :class:`MetricData <htmd.metricdata.MetricData>` object containing the discretized trajectories

    Example
    -------
    >>> model = Model(data)

    .. rubric:: Methods
    .. autoautosummary:: htmd.model.Model
        :methods:
    .. rubric:: Attributes
    .. autoautosummary:: htmd.model.Model
        :attributes:
    """

    def __init__(self, data=None, file=None):
        if data is None:
            if file is not None:
                self.load(file)
            return
        else:
            self.data = data
        self.hmm = None
        self._modelid = None
        if self.data._clusterid is None:
            raise NameError(
                "You need to cluster your data before making a Markov model"
            )
        if self.data._dataid != self.data._clusterid:
            raise NameError(
                "You have modified the data in data.dat after clustering. Please re-cluster."
            )
        self._clusterid = self.data._clusterid

    @staticmethod
    def _get_model(statelist, lagtime, bayesian_samples=None):
        from deeptime.markov.msm import MaximumLikelihoodMSM, BayesianMSM
        from deeptime.markov import TransitionCountEstimator

        count_mode = "sliding" if bayesian_samples is None else "effective"

        counts = TransitionCountEstimator(
            lagtime=lagtime, count_mode=count_mode
        ).fit_fetch(statelist)

        if bayesian_samples is not None:
            return BayesianMSM(n_samples=bayesian_samples).fit_fetch(
                counts.submodel_largest()
            )
        else:
            return MaximumLikelihoodMSM(
                allow_disconnected=False, use_lcc=False
            ).fit_fetch(counts.submodel_largest())

    def markovModel(self, lag, macronum, units="frames", sparse=False):
        """Build a Markov model at a given lag time and calculate metastable states

        Parameters
        ----------
        lag : int
            The lag time at which to calculate the Markov state model. The units are specified with the `units` argument.
        macronum : int
            The number of macrostates (metastable states) to produce
        units : str
            The units of lag. Can be 'frames' or any time unit given as a string.
        sparse : bool
            Make the transition matrix sparse. Useful if lots (> 4000) states are used for the MSM. Warning: untested.

        Examples
        --------
        >>> model = Model(data)
        >>> model.markovModel(150, 4)  # 150 frames lag, 4 macrostates
        """
        self._integrityCheck(markov=True)

        lag = unitconvert(units, "frames", lag, fstep=self.data.fstep)

        statelist = [traj.cluster for traj in self.data.trajectories]
        self.lag = lag
        self.msm = self._get_model(statelist, lag, bayesian_samples=None)
        modelflag = False
        while not modelflag:
            try:
                self.coarsemsm = self.msm.pcca(macronum)
            except Exception as e:
                macronum -= 1
                if macronum < 2:
                    raise RuntimeError(
                        "Could not create even two macrostates. Please revise your clustering."
                    )
                logger.warning(
                    f"PCCA failed with following error. Reducing the number of macrostates to {macronum}. Error: {e}"
                )
                continue

            if len(np.unique(self.coarsemsm.assignments)) != macronum:
                macronum -= 1
                logger.warning(
                    f"PCCA returned empty macrostates. Reducing the number of macrostates to {macronum}."
                )
            else:
                modelflag = True
            if macronum < 2:
                raise RuntimeError(
                    "Could not create even two macrostates. Please revise your clustering."
                )

        self._modelid = random.random()

        logger.info(f"{self.msm.count_fraction * 100:.1f}% of the data was used")

        _macroTrajectoriesReport(
            self.macronum,
            _macroTrajSt(self.data.St, self.macro_ofcluster),
            self.data.simlist,
        )

    def createState(self, microstates=None, indexpairs=None):
        """Creates a new state. Works both for new clusters and macrostates.

        If creating a new cluster, it just reassigns the given frames to the new cluster.
        If creating a new macrostate, it removes the given microstates from their previous macrostate, creates a new one
        and assigns them to it.

        Parameters
        ----------
        microstates : list
            The microstates to split out into a new macrostate.
        indexpairs : list
            List of lists. Each row is a simulation index-frame pair which should be added to a new cluster.
        """
        if microstates is not None and indexpairs is not None:
            raise AttributeError(
                "microstates and indexpairs arguments are mutually exclusive"
            )
        if microstates is not None:
            newmacro = self.macronum

            # Fixing hard assignments
            self.coarsemsm.assignments[microstates] = newmacro

            todelete = np.setdiff1d(np.arange(newmacro), self.coarsemsm.assignments)

            # Fixing memberships. Padding the array with 0s for the new macrostate
            self.coarsemsm._memberships = np.pad(
                self.coarsemsm.memberships,
                ((0, 0), (0, 1)),
                mode="constant",
                constant_values=(0),
            )
            self.coarsemsm._memberships[microstates, :] = 0
            self.coarsemsm._memberships[microstates, -1] = 1

            # Moving probabilities of empty states to new one
            othermicro = np.ones(self.micronum, dtype=bool)
            othermicro[microstates] = False
            othermicro = np.where(othermicro)[0]
            self.coarsemsm._memberships[othermicro, -1] = np.sum(
                self.coarsemsm.memberships[othermicro[:, None], todelete], axis=1
            )

            # Fixing distributions
            self.coarsemsm._metastable_distributions = np.pad(
                self.coarsemsm.metastable_distributions,
                ((0, 1), (0, 0)),
                mode="constant",
                constant_values=(0),
            )
            self.coarsemsm.metastable_distributions[-1, microstates] = 1 / len(
                microstates
            )
        elif indexpairs is not None:
            newcluster = self.data.K
            for ip in indexpairs:
                self.data.trajectories[ip[0]].cluster[ip[1]] = newcluster
            self.data.K += 1
            self.data.N = np.bincount(np.concatenate(self.data.St))

    @property
    def _active_set(self):
        return self.msm.count_model.state_symbols

    @property
    def P(self):
        """The transition probability matrix"""
        return self.msm.transition_matrix

    @property
    def micro_ofcluster(self):
        """Mapping of clusters to microstates

        Numpy array which at index i has the index of the microstate corresponding to cluster i.
        Clusters which were not connected and thus are not in the model have a microstate value of -1.
        """
        self._integrityCheck(postmsm=True)
        micro_ofcluster = -np.ones(self.data.K, dtype=int)
        micro_ofcluster[self._active_set] = np.arange(len(self._active_set))
        return micro_ofcluster

    @property
    def cluster_ofmicro(self):
        """Mapping of microstates to clusters

        Numpy array which at index i has the index of the cluster corresponding to microstate i.
        """
        self._integrityCheck(postmsm=True)
        return self._active_set

    @property
    def micronum(self):
        """Number of microstates"""
        self._integrityCheck(postmsm=True)
        return len(self._active_set)

    @property
    def macronum(self):
        """Number of macrostates"""
        self._integrityCheck(postmsm=True)
        return len(set(self.coarsemsm.assignments))

    @property
    def macro_ofmicro(self):
        """Mapping of microstates to macrostates

        Numpy array which at index i has the index of the macrostate corresponding to microstate i.
        """
        self._integrityCheck(postmsm=True)
        # Fixing deeptime macrostate numbering
        mask = np.ones(np.max(self.coarsemsm.assignments) + 1, dtype=int) * -1
        mask[list(set(self.coarsemsm.assignments))] = range(self.macronum)
        return mask[self.coarsemsm.assignments]

    @property
    def macro_ofcluster(self):
        """Mapping of clusters to microstates

        Numpy array which at index i has the index of the macrostate corresponding to cluster i.
        Clusters which were not connected and thus are not in the model have a macrostate value of -1.
        """
        self._integrityCheck(postmsm=True)
        macro_ofcluster = -np.ones(self.data.K, dtype=int)
        macro_ofcluster[self._active_set] = self.macro_ofmicro
        return macro_ofcluster

    def _plot_implied_timescales(
        self,
        fstep,
        data,
        n_its=None,
        process=None,
        show_mle: bool = True,
        show_samples: bool = True,
        show_sample_mean: bool = True,
        show_sample_confidence: bool = True,
        show_cutoff: bool = True,
        sample_confidence: float = 0.95,
        colors=None,
        ax=None,
        ylog=False,
        **kwargs,
    ):
        r"""Creates an implied timescales plot inside exising matplotlib axes.

        ! Taken from deeptime !
        .. plot:: examples/plot_implied_timescales.py

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            The matplotlib axes to use for plotting.
        data : ImpliedTimescales
            A timescales data container object, can be obtained, e.g., via :meth:`ImpliedTimescales.from_models`.
        n_its : int, optional, default=None
            Maximum number of timescales to plot.
        process : int, optional, default=None
            A particular process to plot. This is mutually exclusive with n_its.
        show_mle : bool, default=True
            Whether to show the timescale of the maximum-likelihood estimate.
        show_samples : bool, default=True
            Whether to show sample means and/or confidences.
        show_sample_mean : bool, default=True
            Whether to show the sample mean. Only has an effect if show_samples is True and there are samples in the data.
        show_sample_confidence : bool, default=True
            Whether to show the sample confidence. Only has an effect if show_samples is True and there are samples
            in the data.
        show_cutoff : bool, default=True
            Whether to show the model resolution cutoff as grey filled area.
        sample_confidence : float, default=0.95
            The confidence to plot. The default amounts to a shaded area containing 95% of the sampled values.
        colors : list of colors, optional, default=None
            The colors that should be used for timescales. By default uses the matplotlib default colors as per
            rc-config value "axes.prop_cycle".
        **kwargs
            Keyword arguments which are forwarded into the matplotlib plotting function for timescales.

        Returns
        -------
        ax : matplotlib.axes.Axes
            The matplotlib axes that were used to plot the timescales.

        See Also
        --------
        ImpliedTimescales
        """
        from deeptime.plots.util import default_colors
        from deeptime.util import confidence_interval

        if ax is None:
            import matplotlib.pyplot as plt

            ax = plt.gca()
        if n_its is not None and process is not None:
            raise ValueError("n_its and process are mutually exclusive.")
        if process is not None and process >= data.max_n_processes:
            raise ValueError(
                f"Requested process {process} when only {data.max_n_processes} are available."
            )

        if process is None and n_its is None:
            n_its = data.max_n_processes
        it_indices = [process] if process is not None else np.arange(n_its)
        if colors is None:
            colors = default_colors()
        for it_index in it_indices:
            color = colors[it_index % len(colors)]
            if show_mle:
                ax.plot(
                    data.lagtimes * fstep,
                    data.timescales_for_process(it_index) * fstep,
                    color=color,
                    **kwargs,
                )
            if data.has_samples and show_samples:
                its_samples = data.samples_for_process(it_index)
                if show_sample_mean:
                    sample_mean = np.nanmean(its_samples, axis=1)
                    ax.plot(
                        data.lagtimes * fstep,
                        sample_mean * fstep,
                        marker="o",
                        linestyle="dashed",
                        color=color,
                    )
                if show_sample_confidence:
                    l_conf, r_conf = confidence_interval(
                        its_samples.T, conf=sample_confidence, remove_nans=True
                    )
                    ax.fill_between(
                        data.lagtimes * fstep,
                        l_conf * fstep,
                        r_conf * fstep,
                        alpha=0.2,
                        color=color,
                    )

        if show_cutoff:
            ax.plot(
                data.lagtimes * fstep, data.lagtimes * fstep, linewidth=2, color="black"
            )
            ax.fill_between(
                data.lagtimes * fstep,
                np.full((data.n_lagtimes,), fill_value=ax.get_ylim()[0]) * fstep,
                data.lagtimes * fstep,
                alpha=0.5,
                color="grey",
            )
        if ylog:
            ax.set_yscale("log")

        ax.set_title("Implied timescales")
        ax.set_xlabel("Lag time (ns)")
        ax.set_ylabel("Timescale (ns)")
        return ax

    def plotTimescales(
        self,
        lags=None,
        minlag=None,
        maxlag=None,
        numlags=25,
        units="frames",
        errors=None,
        nits=None,
        results=False,
        plot=True,
        save=None,
        njobs=1,
        ylog=True,
    ):
        """Plot the implied timescales of MSMs of various lag times

        Parameters
        ----------
        lags : list
            Specify explicitly at which lag times to compute the timescales.
        minlag: float
            The minimum lag time for the timescales. Used in combination with `maxlag` and `numlags`.
        maxlag: float
            The maximum lag time for the timescales. If None will default to the mode length of the trajectories.
        numlags: int
            The number of points to place between `minlag` and `maxlag`.
        units : str
            The units of lag. Can be 'frames' or any time unit given as a string.
        errors : errors
            Calculate errors using Bayesian MSMs
        nits : int
            Number of implied timescales to calculate. Default: all
        results : bool
            If the method should return the calculated implied timescales
        plot : bool
            If the method should display the plot of implied timescales
        save : str
            Path of the file in which to save the figure
        njobs : int
            DEPRECATED: Number of parallel jobs to spawn for calculation of timescales. Negative numbers are used for spawning jobs as many as CPU threads.
            -1: for all CPUs -2: for all except one etc.
        ylog : bool
            Set to False to get linear y axis instead of logarithmic

        Returns
        -------
        If given results=True this method will return the following data
        its : np.ndarray
            The calculated implied timescales. 2D array with dimensions (len(`lags`), `nits`)
        lags : np.ndarray
            A list of the lag times that were used to calculate the implied timescales

        Examples
        --------
        >>> model = Model(data)
        >>> model.plotTimescales()
        >>> model.plotTimescales(lags=list(range(1,100,5)))
        >>> model.plotTimescales(minlag=0.1, maxlag=20, numlags=25, units='ns')
        """
        from deeptime.util.validation import implied_timescales
        from tqdm import tqdm

        self._integrityCheck()
        if lags is None:
            lags = self.data._defaultLags(minlag, maxlag, numlags, units)
        else:
            lags = unitconvert(units, "frames", lags, fstep=self.data.fstep).tolist()

        if nits is None:
            nits = np.min((self.data.K - 1, 20))

        statelist = [traj.cluster for traj in self.data.trajectories]
        models = []
        for lagtime in tqdm(lags, desc="Estimating Timescales"):
            models.append(self._get_model(statelist, lagtime, bayesian_samples=errors))

        its_data = implied_timescales(models, n_its=nits)
        if plot or (save is not None):
            from matplotlib import pylab as plt

            if plot:
                plt.ion()
            plt.figure()
            ax = plt.gca()
            try:
                self._plot_implied_timescales(
                    self.data.fstep, data=its_data, n_its=nits, ax=ax, ylog=ylog
                )
            except ValueError as ve:
                plt.close()
                raise ValueError(
                    f"{ve} This is probably caused by badly set fstep in the data ({self.data.fstep}). "
                    + "Please correct the model.data.fstep to correspond to the simulation frame step in nanoseconds."
                )
            if save is not None:
                plt.savefig(save, dpi=300, bbox_inches="tight", pad_inches=0.2)
            if plot:
                plt.show()
        if results:
            return its_data._its, its_data.lagtimes

    def maxConnectedLag(self, lags, njobs=1):
        """Heuristic for getting the lagtime before a timescale drops.

        It calculates the last lagtime before a drop occurs in the first implied timescale due to disconnected states.
        If the top timescale is closer to the second top timescale at the previous lagtime than to itself at the previous
        lagtime it means that a drop occured. The lagtime before the drop is returned.

        Parameters
        ----------
        lags : np.ndarray or list
            A list of lag times for which to calculate the implied timescales

        Returns
        -------
        ml : int
            The maximum lagtime before a drop occurs in the top timescale

        Examples
        --------
        >>> model = Model(data)
        >>> model.maxConnectedLag(list(range(1, 100, 5)))
        """
        from deeptime.util.validation import implied_timescales
        from tqdm import tqdm

        if len(lags) == 1:
            return lags
        if isinstance(lags, np.ndarray):
            lags = lags.astype(int)

        statelist = [traj.cluster for traj in self.data.trajectories]
        models = []
        for lagtime in tqdm(lags, desc="Estimating Timescales"):
            models.append(self._get_model(statelist, lagtime, bayesian_samples=None))

        its_data = implied_timescales(models, n_its=2)
        itime = its_data._its

        for i in range(1, np.size(itime, 0)):
            if abs(itime[i, 0] - itime[i - 1, 1]) < abs(itime[i, 0] - itime[i - 1, 0]):
                lagidx = i - 1
                break
            else:
                lagidx = i
        return lags[lagidx], itime

    def sampleStates(
        self,
        states=None,
        frames=20,
        statetype="macro",
        replacement=False,
        samplemode="random",
        allframes=False,
    ):
        """Samples frames from a set of states

        Parameters
        ----------
        states : Union[list, int]
            A list of state indexes from which we want to sample
        frames : Union[None, int, list]
            An integer with the number of frames we want to sample per state or a list of same length as
            `states` which contains the number of frames we want from each of the states.
            If set to None it will return all frames of the states.
        statetype : ['micro','macro','cluster'], optional
            The type of state we want to sample from.
        replacement : bool
            If we want to sample with or without replacement.
        samplemode : ['random','even','weighted']
            What sampling strategy to use. For `statetype` == 'macro' this can be set to 'even' to sample evenly from
            all microstates in the macrostate or to 'weighted' to sample proportional to the equilibium probability of
            each microstate in the macrostate.
        allframes : bool
            Deprecated. Use frames=None instead.

        Returns
        -------
        absframes : numpy.ndarray
            An array which contains for each state an array containing absolute trajectory frames
        relframes : numpy.ndarray
            An array which contains for each state a 2D array containing the trajectory ID and frame number for each of
            the sampled frames

        Examples
        --------
        >>> model = Model(data)
        >>> model.markovModel(100, 5)
        >>> model.sampleStates(range(5), [10, 3, 2, 50, 1])  # Sample from all 5 macrostates
        >>> model.sampleStates(range(model.micronum), samplesnum, statetype='micro')  # Sample from all microstates
        """
        if statetype == "cluster":
            if samplemode != "random":
                logger.warning(
                    "'cluster' states incompatible with 'samplemode' other than 'random'. Defaulting to 'random'"
                )
            return self.data.sampleClusters(states, frames, replacement, allframes)

        if states is None:
            if statetype == "macro":
                states = range(self.macronum)
            elif statetype == "micro":
                states = range(self.micronum)
        if isinstance(states, int):
            states = [
                states,
            ]

        if allframes:
            logger.warning(
                "The allframes option will be deprecated. Use frames=None instead."
            )
            frames = None

        self._integrityCheck(postmsm=True)
        if statetype != "macro" and samplemode != "random":
            samplemode = "random"
            logger.warning(
                "'micro' states incompatible with 'samplemode' other than 'random'. Defaulting to 'random'"
            )

        if frames is None or isinstance(frames, int):
            frames = np.repeat(frames, len(states))

        stConcat = np.concatenate(self.data.St)
        absFrames = []
        relFrames = []
        for i in range(len(states)):
            if frames[i] == 0:
                absFrames.append(np.array([], dtype=int))
                relFrames.append(np.array([], dtype=int))
                continue

            st = states[i]
            if statetype == "macro":
                (selFr, selMicro) = _sampleMacro(
                    self, st, stConcat, samplemode, frames[i], replacement
                )
                absFrames.append(selFr)
            elif statetype == "micro":
                absFrames.append(
                    _sampleMicro(self, st, stConcat, frames[i], replacement)
                )
            else:
                raise NameError("No valid state type given (read documentation)")

            if len(absFrames[-1]) == 0:
                raise NameError(
                    "No frames could be sampled from {} state {}. State is empty.".format(
                        statetype, st
                    )
                )

            relFrames.append(self.data.abs2rel(absFrames[-1]))
        return absFrames, relFrames

    def eqDistribution(self, plot=True, save=None):
        """Obtain and plot the equilibrium probabilities of each macrostate

        Parameters
        ----------
        plot : bool, optional, default=True
            Disable plotting of the equilibrium distribution by setting it to False
        save : str
            Path of the file in which to save the figure

        Returns
        -------
        eq : ndarray
            An array of equilibrium probabilities of the macrostates

        Examples
        --------
        >>> model = Model(data)
        >>> model.markovModel(100, 5)
        >>> model.eqDistribution()
        """
        # logger.warning('Equilibrium distribution calculations for macrostates are now done using membership '
        #                'probabilities and hence your results might differ from analyses done before this change.')
        self._integrityCheck(postmsm=True)
        macroeq = np.ones(self.macronum) * -1
        macroindexes = list(set(self.coarsemsm.assignments))
        for i in range(self.macronum):
            # macroeq[i] = np.sum(self.msm.stationary_distribution[self.macro_ofmicro == i])
            macroeq[i] = np.sum(
                self.coarsemsm.memberships[:, macroindexes[i]]
                * self.msm.stationary_distribution
            )

        if plot or (save is not None):
            from matplotlib import pylab as plt

            if plot:
                plt.ion()
            plt.figure()
            plt.bar(np.arange(self.macronum) + 0.4, macroeq)
            plt.ylabel("Equilibrium probability")
            plt.xlabel("Macrostates")
            plt.xticks(np.arange(self.macronum) + 0.4, range(self.macronum))
            if save is not None:
                plt.savefig(save, dpi=300, bbox_inches="tight", pad_inches=0.2)
            if plot:
                plt.show()

        return macroeq

    def _coarseP(self):
        M = self.coarsemsm.memberships
        Pcoarse = np.linalg.inv(M.T.dot(M)).dot(M.T).dot(self.P).dot(M)
        if len(np.where(Pcoarse < 0)[0]) != 0:
            raise NameError(
                "Cannot produce coarse P matrix. Ended up with negative probabilities. Try using less macrostates."
            )
        return Pcoarse

    def getStates(
        self,
        states=None,
        statetype="macro",
        wrapsel="protein",
        alignsel="name CA",
        alignmol=None,
        samplemode="weighted",
        numsamples=50,
        simlist=None,
    ):
        """Get samples of MSM states in Molecule classes

        Parameters
        ----------
        states : ndarray, optional
            A list of states to visualize
        statetype : ['macro','micro','cluster'], optional
            The type of state to visualize
        wrapsel : str, optional, default='protein'
            Atom selection string to use for wrapping.
            See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
        alignsel : str, optional, default='name CA'
            Atom selection string used for aligning all frames. Set to None to disable aligning.
            See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
        alignmol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A reference molecule onto which to align all others
        samplemode : ['weighted','random'], optional, default='weighted'
            How to obtain the samples from the states
        numsamples : int
            Number of samples (conformations) for each state.
        simlist : numpy.ndarray of :class:`Sim <htmd.simlist.Sim>` objects
            Optionally pass a different (but matching, i.e. filtered) simlist for creating the Molecules.

        Returns
        -------
        mols : ndarray of :class:`Molecule <moleculekit.molecule.Molecule>` objects
            A list of :class:`Molecule <moleculekit.molecule.Molecule>` objects containing the samples of each state

        Examples
        --------
        >>> model = Model(data)
        >>> model.markovModel(100, 5)
        >>> mols = model.getStates()
        >>> for m in mols:
        >>>     m.view()
        """
        from htmd.projections.metric import _singleMolfile

        self._integrityCheck(postmsm=(statetype != "cluster"))
        if simlist is None:
            simlist = self.data.simlist
        else:
            if len(simlist) != len(self.data.simlist):
                raise AttributeError(
                    "Provided simlist has different number of trajectories than the one used by the model."
                )

        (single, molfile) = _singleMolfile(simlist)
        if not single:
            raise NameError(
                "Visualizer does not support yet visualization of systems with different structure files. "
                "The simlist should be created with a single molfile (for example a filtered one)"
            )
        if alignmol is None:
            alignmol = Molecule(molfile)
        if statetype != "macro" and statetype != "micro" and statetype != "cluster":
            raise NameError("'statetype' must be either 'macro', 'micro' or ''cluster'")
        if states is None:
            if statetype == "macro":
                states = range(self.macronum)
            elif statetype == "micro":
                states = range(self.micronum)
            elif statetype == "cluster":
                states = range(self.data.K)
        if len(states) == 0:
            raise NameError("No " + statetype + " states exist in the model")

        (tmp, relframes) = self.sampleStates(
            states, numsamples, statetype=statetype, samplemode=samplemode
        )

        from tqdm import tqdm

        # This loop really iterates over states. sampleStates returns an array of arrays
        mols = []
        for rel in tqdm(relframes, desc="Getting state Molecules"):
            mols.append(
                self._loadMols(rel, molfile, wrapsel, alignsel, alignmol, simlist)
            )

        return mols

    def viewStates(
        self,
        states=None,
        statetype="macro",
        protein=None,
        ligand=None,
        mols=None,
        numsamples=50,
        wrapsel="protein",
        alignsel="name CA",
        gui=False,
        simlist=None,
    ):
        """Visualize macro/micro/cluster states in VMD

        Parameters
        ----------
        states : ndarray, optional
            A list of states to visualize
        statetype : ['macro','micro','cluster'], optional
            The type of state to visualize
        protein : bool, optional
            Set to True to enable pure protein system visualization
        ligand : str, optional
            Atom selection string for the ligand.
            See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
        mols : ndarray, optional
            An array of :class:`Molecule <moleculekit.molecule.Molecule>` objects to visualize
        numsamples : int
            Number of samples (conformations) for each state.
        wrapsel : str, optional, default='protein'
            Atom selection string to use for wrapping.
            See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
        alignsel : str, optional, default='name CA'
            Atom selection string used for aligning all frames. See to None to disable aligning.
            See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
        simlist : numpy.ndarray of :class:`Sim <htmd.simlist.Sim>` objects
            Optionally pass a different (but matching, i.e. filtered) simlist for visualizing the states.

        Examples
        --------
        >>> model = Model(data)
        >>> model.markovModel(100, 5)
        >>> model.viewStates(protein=True)

        >>> model.viewStates(ligand='resname MOL')
        """
        from moleculekit.config import _config

        self._integrityCheck(postmsm=(statetype != "cluster"))

        if _config["viewer"] is not None and _config["viewer"].lower() in (
            "ngl",
            "webgl",
        ):
            if protein is None and ligand is None:
                raise RuntimeError(
                    "You need to specify either the `protein` of `ligand` arguments for state visualization."
                )

            return self._viewStatesNGL(
                states, statetype, protein, ligand, mols, numsamples, gui=gui
            )

        if states is None:
            states = range(self.macronum)
        if isinstance(states, int):
            states = [states]
        if mols is None:
            mols = self.getStates(
                states,
                statetype,
                numsamples=numsamples,
                wrapsel=wrapsel,
                alignsel=alignsel,
                simlist=simlist,
            )
        colors = [0, 1, 3, 4, 5, 6, 7, 9]
        for i, s in enumerate(states):
            mols[i].viewname = f"{statetype} {s}"
            mols[i].reps.add(
                sel="protein or nucleic",
                style="NewCartoon",
                frames=range(mols[i].numFrames) if ligand is None else None,
            )
            if ligand is not None:
                mols[i].reps.add(
                    sel=ligand,
                    style="Lines",
                    color=colors[np.mod(i, len(colors))],
                    frames=range(mols[i].numFrames),
                )
            mols[i].view()

    def _viewStatesNGL(
        self, states, statetype, protein, ligand, mols, numsamples, gui=False
    ):
        from moleculekit.util import sequenceID

        if states is None:
            states = range(self.macronum)
        if isinstance(states, int):
            states = [states]
        if mols is None:
            mols = self.getStates(states, statetype, numsamples=min(numsamples, 15))
        hexcolors = {
            0: "#0000ff",
            1: "#ff0000",
            2: "#333333",
            3: "#ff6600",
            4: "#ffff00",
            5: "#4c4d00",
            6: "#b2b2cc",
            7: "#33cc33",
            8: "#ffffff",
            9: "#ff3399",
            10: "#33ccff",
        }
        if protein is None and ligand is None:
            raise NameError(
                'Please provide either the "protein" or "ligand" parameter for viewStates.'
            )
        k = 0
        from nglview import NGLWidget, HTMDTrajectory

        view = NGLWidget(gui=gui)
        ref = mols[0].copy()
        for i, s in enumerate(states):
            if protein:
                mol = Molecule()
            if ligand:
                mol = ref.copy()
                mol.remove(ligand, _logger=False)
                mol.dropFrames(keep=0)
                mols[i].filter(ligand, _logger=False)
            mols[i].set("chain", "{}".format(s))
            tmpcoo = mols[i].coords
            for j in range(mols[i].numFrames):
                mols[i].coords = np.atleast_3d(tmpcoo[:, :, j])
                if ligand:
                    mols[i].set("segid", sequenceID(mols[i].resid) + k)
                    k = int(mols[i].segid[-1])
                mol.append(mols[i])
            view.add_trajectory(HTMDTrajectory(mol))
            # Setting up representations
            if ligand:
                view[i].add_cartoon("protein", color="sstruc")
                view[i].add_hyperball(
                    ":{}".format(s), color=hexcolors[np.mod(i, len(hexcolors))]
                )
            if protein:
                view[i].add_cartoon("protein", color="residueindex")

        self._nglButtons(view, statetype, states)
        return view

    def _nglButtons(self, ngl_widget, statetype, states):
        # Adds buttons for enabling and disabling macrostate visualizations
        import ipywidgets
        from IPython.display import display

        container = []
        for s in states:
            w = ipywidgets.Checkbox(description="{} {}".format(statetype, s))
            w.value = True
            container.append(w)

        def updateReps(name):
            on = []
            for i, w in enumerate(container):
                if w.value:
                    on.append(i)
            ngl_widget.show_only(on)

        for w in container:
            w.on_trait_change(updateReps, "value")

        # container.append(ipywidgets.Checkbox(description="all"))

        hb = ipywidgets.HBox(container)
        display(hb)

    def save(self, filename):
        """Save a :class:`Model <htmd.model.Model>` object to disk

        Parameters
        ----------
        filename : str
            Path of the file in which to save the object

        Examples
        --------
        >>> model = Model(data)
        >>> model.markovModel(100, 5)
        >>> model.save('./model.dat')
        """
        import pickle

        # Temporarily store data object and replace with dicts
        tmpdata = self.data
        if self.data.parent is not None:
            tmpparentdata = self.data.parent
            self.data.parent = self.data.parent.__dict__
        self.data = self.data.__dict__

        # Dump the dict
        f = open(filename, "wb")
        pickle.dump(self.__dict__, f)
        f.close()

        # Restore data to classes
        self.data = tmpdata
        if self.data.parent is not None:
            self.data.parent = tmpparentdata

    def load(self, filename):
        """Load a :class:`MetricData <htmd.metricdata.MetricData>` object from disk

        Parameters
        ----------
        filename : str
            Path to the saved MetricData object

        Examples
        --------
        >>> model = Model()
        >>> model.load('./model.dat')
        """
        import sys
        import pickle
        from htmd.metricdata import MetricData

        try:
            import pandas.indexes
        except ImportError:
            import pandas.core.indexes

            sys.modules["pandas.indexes"] = (
                pandas.core.indexes
            )  # Hacky fix for new pandas version

        f = open(filename, "rb")
        z = pickle.load(f)
        f.close()
        for k in z:
            if k == "data":
                m = MetricData()
                m.load(z[k])
                self.__dict__[k] = m
            else:
                self.__dict__[k] = z[k]

    def copy(self):
        """Produces a deep copy of the object

        Returns
        -------
        data : :class:`Model` object
            A copy of the current object

        Examples
        --------
        >>> model2 = model.copy()
        """
        from copy import deepcopy

        return deepcopy(self)

    def cktest(
        self,
        lags=None,
        minlag=None,
        maxlag=None,
        numlags=25,
        units="frames",
        plot=True,
        save=None,
        errors=None,
    ):
        """Conducts a Chapman-Kolmogorov test.

        Parameters
        ----------
        lags: list, optional
            List of lag times to test
        minlag : int, optional
            Minimum lag time to test. Used if lags is None. By default it will
            use 10 if maxlag is greater than 20, otherwise it will use 2.
        maxlag : int, optional
            Maximum lag time to test. Used if lags is None. By default it will
            use the mode of the trajectory lengths.
        numlags : int, optional
            Number of lag times to test between minlag and maxlag. Used if lags is None.
        units : str, optional
            Units of the lag times. By default it will use frames.
        plot : bool
            If the method should display the plot of the CK test
        save : str
            Path of the file in which to save the figure
        errors : int, optional
            Number of Bayesian samples to use for the error bars. If set to None,
            it will not plot error bars and use a Maximum Likelihood MSM.
        """
        from deeptime.plots.chapman_kolmogorov import plot_ck_test
        from matplotlib import pylab as plt

        if lags is None:
            lags = self.data._defaultLags(minlag, maxlag, numlags, units)
        else:
            lags = unitconvert(units, "frames", lags, fstep=self.data.fstep).tolist()

        statelist = [traj.cluster for traj in self.data.trajectories]
        models = []
        for lag in lags:
            models.append(self._get_model(statelist, lag, bayesian_samples=errors))

        res = models[0].ck_test(models, n_metastable_sets=self.macronum)
        plot_ck_test(res, legend=True)

        if save is not None:
            plt.savefig(save, dpi=300, bbox_inches="tight", pad_inches=0.2)
        if plot:
            plt.show()

    def createCoreSetModel(self, threshold=0.5):
        """Given an MSM this function detects the states belonging to a core set and returns a new model consisting
        only of these states.

        Parameters
        ----------
        threshold : float
            Membership probability threshold over which microstates belong to the core of a macrostate

        Returns
        -------
        newmodel :
            A new model object
        frames : list
            A list of the frames that were kept in the new model
        """
        if (threshold >= 1) or (threshold <= 0):
            raise AttributeError(
                "threshold argument only accepts values between (0, 1)"
            )

        def calcCoreSet(distr, assign, threshold):
            coreset = []
            for i, md in enumerate(distr):
                microofmacro = np.where(assign == i)[0]
                prob = md[microofmacro]
                tt = threshold * (prob.max() - prob.min())
                coreset.append(microofmacro[np.where(prob > tt)[0]])
            return coreset

        def coreDtraj(data, micro_ofcluster, coreset):
            corestates = np.concatenate(coreset)
            newmapping = np.ones(corestates.max() + 1, dtype=int) * -1
            newmapping[corestates] = np.arange(len(corestates))
            discretetraj = [st.copy() for st in data.St]
            frames = []
            newdiscretetraj = []
            newcounts = np.zeros(len(corestates))
            for t, st in enumerate(discretetraj):
                oldmicro = None
                newtraj = []
                tframes = []
                for f, cl in enumerate(st):
                    newmicro = None
                    micro = micro_ofcluster[cl]
                    if micro == -1:  # If we are in an dropped cluster, keep old index
                        newmicro = oldmicro
                    else:
                        for co in coreset:
                            if micro in co:
                                newmicro = micro
                                oldmicro = micro
                                break
                    if newmicro is None and oldmicro is not None:
                        newtraj.append(oldmicro)
                        tframes.append(f)
                    elif newmicro is not None:
                        newtraj.append(newmicro)
                        tframes.append(f)
                mappedtraj = newmapping[np.array(newtraj, dtype=int)]
                newdiscretetraj.append(mappedtraj)
                if len(mappedtraj):
                    newcounts[: mappedtraj.max() + 1] += np.bincount(mappedtraj)
                frames.append(np.array(tframes))
            # kept = np.array([i for i, x in enumerate(newdiscretetraj) if len(x) != 0])
            return (
                newdiscretetraj,
                len(corestates),
                newcounts,
                frames,
            )

        coreset = calcCoreSet(
            self.coarsemsm.metastable_distributions,
            self.coarsemsm.assignments,
            threshold,
        )
        newdata = self.data.copy()
        newSt, newdata.K, newdata.N, frames = coreDtraj(
            self.data, self.micro_ofcluster, coreset
        )
        for i, (s, fr) in enumerate(zip(newSt, frames)):
            if len(s):
                newdata.trajectories[i].cluster[fr] = s

        logger.info(
            f"Kept {[len(x) for x in coreset]} microstates from each macrostate."
        )

        dataobjects = [newdata]
        if newdata.parent is not None:
            dataobjects.append(newdata.parent)
        for data in dataobjects:
            dat = []
            ref = []
            simstmp = []
            cluster = []
            for i, fr in enumerate(frames):
                if len(fr):
                    dat.append(data.trajectories[i].projection[fr, :])
                    ref.append(data.trajectories[i].reference[fr, :])
                    if len(data.trajectories[i].cluster):
                        cluster.append(data.trajectories[i].cluster[fr])
                    simstmp.append(data.trajectories[i].sim)

            data._loadTrajectories(dat, ref, simstmp, cluster if len(cluster) else None)

        return Model(newdata), frames

    def _getFEShistogramCounts(
        self,
        data,
        dimx,
        dimy,
        nbins=80,
        pad=0.5,
        micro_ofcluster=None,
        stationary_distribution=None,
    ):
        from tqdm import tqdm

        clusters = np.hstack(data.St)
        data_dim = []
        for traj in data.trajectories:
            data_dim.append(traj.projection[:, [dimx, dimy]])
        data_dim = np.vstack(data_dim)

        xmin, ymin = data_dim.min(axis=0)
        xmax, ymax = data_dim.max(axis=0)
        hist_range = np.array([[xmin - pad, xmax + pad], [ymin - pad, ymax + pad]])

        if micro_ofcluster is None:
            counts, xbins, ybins = np.histogram2d(
                data_dim[:, 0], data_dim[:, 1], bins=nbins, range=hist_range
            )
            return counts.T, xbins, ybins

        counts = np.zeros((nbins, nbins))
        for m in tqdm(range(len(stationary_distribution))):
            frames = micro_ofcluster[clusters] == m
            if frames.sum():
                state_mask, xbins, ybins = np.histogram2d(
                    data_dim[frames, 0],
                    data_dim[frames, 1],
                    bins=nbins,
                    range=hist_range,
                )
                state_mask = state_mask.T > 0
                counts += state_mask * stationary_distribution[m]
        return counts, xbins, ybins

    def plotFES(
        self,
        dimX,
        dimY,
        temperature,
        states=False,
        s=10,
        cmap=None,
        fescmap=None,
        statescmap=None,
        plot=True,
        save=None,
        data=None,
        levels=7,
    ):
        """Plots the free energy surface on any given two dimensions. Can also plot positions of states on top.

        Parameters
        ----------
        dimX : int
            Index of projected dimension to use for the X axis.
        dimY : int
            Index of projected dimension to use for the Y axis.
        temperature : float
            Simulation temperature.
        states : bool
            If True, will plot scatter plot of microstates coloured by macro state on top of FES.
        s : float
            Marker size for states.
        cmap :
            Sets the Matplotlib colormap for both `fescmap` and `statescmap`
        fescmap :
            Matplotlib colormap for the free energy surface
        statescmap:
            Matplotlib colormap for the states
        plot : bool
            If the method should display the plot of the FES. If both plot=False and save=None, the method will return the figure and axes.
        save : str
            Path of the file in which to save the figure. If both plot=False and save=None, the method will return the figure and axes.
        data : :class:`MetricData` object
            Optionally you can pass a different MetricData object than the one used to build the model. For example
            if the user wants to build a model on distances but wants to plot the FES on top of RMSD values. The
            MetricData object needs to have the same simlist as the Model.

        Returns
        -------
        f : matplotlib.figure.Figure
            The matplotlib figure object
        ax : matplotlib.axes.Axes
            The matplotlib axes object

        Examples
        --------
        >>> import matplotlib as plt
        >>> model.plotFES(0, 1, 300)
        >>> model.plotFES(2, 3, 300, data=otherdata, states=True, fescmap=plt.cm.gray)
        """
        self._integrityCheck(postmsm=True)
        from matplotlib import pylab as plt
        from htmd.kinetics import Kinetics

        if data is None:
            data = self.data
            microcenters = self.data.Centers[self.cluster_ofmicro, :]
        else:
            if self.data.numFrames != data.numFrames or ~np.all(
                [s1 == s2 for s1, s2 in zip(self.data.simlist, data.simlist)]
            ):
                raise RuntimeError(
                    "The data argument you provided uses a different simlist than the Model."
                )
            microcenters = np.vstack(
                getStateStatistic(self, data, range(self.micronum), statetype="micro")
            )

        if fescmap is None:
            fescmap = "viridis"
        if statescmap is None:
            statescmap = plt.cm.jet
        if cmap is not None:
            fescmap = cmap
            statescmap = cmap

        xlabel = f"Dimension {dimX}"
        if data.description is not None:
            xlabel = data.description.description[dimX]

        ylabel = f"Dimension {dimY}"
        if data.description is not None:
            ylabel = data.description.description[dimY]

        title = "Free energy surface"

        counts, xbins, ybins = data._getFEShistogramCounts(
            dimX,
            dimY,
            nbins=80,
            pad=0.5,
            micro_ofcluster=self.micro_ofcluster,
            stationary_distribution=self.msm.stationary_distribution,
            St=self.data.St,
        )
        counts /= counts.sum()  # Normalize probabilites
        nonzero = counts != 0
        energy = counts.copy()
        energy[nonzero] = -Kinetics._kB * temperature * np.log(counts[nonzero])
        energy[~nonzero] = energy[nonzero].max() + 100
        f, ax, cf = data._contourPlot(
            energy,
            xbins,
            ybins,
            levels=levels,
            nonzero=nonzero,
            cmap=fescmap,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
        )

        data._setColorbar(f, cf, "kcal/mol", scientific=False)
        if states:
            colors = statescmap(np.linspace(0, 1, self.macronum))
            for m in range(self.macronum):
                macromicro = microcenters[self.macro_ofmicro == m, :]
                _ = ax.scatter(
                    macromicro[:, dimX],
                    macromicro[:, dimY],
                    s=s,
                    color=colors[m],
                    label=f"Macro {m}",
                    edgecolors="none",
                )
            ax.legend(prop={"size": 8})

        if save is not None:
            plt.savefig(save, dpi=300, bbox_inches="tight", pad_inches=0.2)
            plt.close()
        elif plot:
            plt.show()
        else:
            return f, ax

    def _integrityCheck(self, postmsm=False, markov=False):
        if postmsm and self._modelid is None:
            raise NameError("You need to call markovModel before calling this command")
        if (
            not markov
            and self.data._dataid == self.data._clusterid
            and self.data._dataid != self._clusterid
        ):
            raise NameError(
                "After updating the MetricData object you need to call the markovModel command anew."
            )
        if self.data._dataid != self.data._clusterid:
            raise NameError(
                "After modifying the data in the MetricData object you need to recluster and reconstruct the markov model."
            )

    def _loadMols(self, rel, molfile, wrapsel, alignsel, refmol, simlist):
        frames = self.data.rel2sim(rel, simlist=simlist)
        mol = Molecule(molfile)
        trajs = []
        frs = []
        for f in frames:
            trajs.append(f.sim.trajectory[f.piece])
            frs.append(f.frame)
        mol.read(np.array(trajs), frames=np.array(frs))
        if wrapsel is not None and len(wrapsel):
            mol.wrap(wrapsel)
        if alignsel is not None and len(alignsel):
            mol.align(alignsel, refmol=refmol)
        return mol


def getStateStatistic(
    reference,
    data,
    states,
    statetype="macro",
    weighted=False,
    method=np.mean,
    axis=0,
):
    """Calculates properties of the states.

    Calculates properties of data corresponding to states. Can calculate for example the mean distances of atoms in a
    state, or the standard deviation of the RMSD in a state.

    Parameters
    ----------
    reference : :class:`Model` object or :class:`MetricData` object
        A model containing the state definitions or a MetricData object containing cluster definitions
    data : :class:`MetricData` object
        A projection corresponding to the conformations in the states of the model. The data and the model need to share
        the same `simlist`
    states : list
        A list of the states for which we want to calculate the properties
    statetype : ['macro','micro','cluster']
        The state type
    weighted : bool
        If the properties of the macrostates should be calculated as the weighted average of their microstates, where the
        weights are the equilibrium probabilities of the microstates
    method :
        A function pointer for the method that calculates our desired property. e.g. np.mean, np.std
    axis : int
        On which axis of the data the method should operate on

    Returns
    -------
    statistic : list
        A list which contains in each element the desired statistic of a specific state specifies in `states`

    Examples
    --------
    >>> data = MetricDistance.project(sims, 'protein and name CA', 'resname MOL')
    >>> model = Model(data)
    >>> model.markovModel(100, 5)
    >>> # Get the standard deviation of distances in all macrostates
    >>> getStateStatistic(model, data, list(range(5)), method=np.std)
    """
    from htmd.metricdata import MetricData

    if axis != 0:
        logger.warning("Axis different than 0 might not work correctly yet")

    if isinstance(reference, Model):
        refdata = reference.data
    elif isinstance(reference, MetricData):
        refdata = reference
        if statetype != "cluster":
            raise RuntimeError(
                "You can only use statetype cluster with reference MetricData object. To use other statetypes build a Model."
            )
    else:
        raise RuntimeError("Invalid argument type")

    if refdata.numTrajectories > 0 and np.any(refdata.trajLengths != data.trajLengths):
        raise NameError(
            "Data trajectories need to match in size and number to the trajectories in the model"
        )
    stconcat = np.concatenate(refdata.St)
    datconcat = np.concatenate(data.dat)

    statistic = []
    for i, st in enumerate(states):
        if statetype == "macro":
            frames = reference.macro_ofcluster[stconcat] == st
        elif statetype == "micro":
            frames = reference.micro_ofcluster[stconcat] == st
        elif statetype == "cluster":
            frames = stconcat == st
        else:
            raise NameError("No valid state type given (read documentation)")

        if statetype == "macro" and weighted:
            statistic.append(
                _weightedMethod(reference, method, stconcat, datconcat, st, axis)
            )
        else:
            if axis is None:
                statistic.append(method(datconcat[frames, ...]))
            else:
                statistic.append(method(datconcat[frames, ...], axis=axis))
    return statistic


def _weightedMethod(model, method, stconcat, datconcat, st, axis):
    microsofmacro = np.where(model.macro_ofmicro == st)[0]
    eq = model.msm.stationary_distribution
    weights = eq / np.sum(eq[microsofmacro])
    avgstatistic = np.zeros(np.size(datconcat, 1))
    for m in microsofmacro:
        frames = model.micro_ofcluster[stconcat] == m
        if axis is None:
            stat = method(datconcat[frames, :])
        else:
            stat = method(datconcat[frames, :], axis=axis)
        avgstatistic = avgstatistic + stat * weights[m]
    return avgstatistic


def macroAccumulate(model, microvalue):
    """Accumulate values of macrostates from a microstate array

    Parameters
    ----------
    model : :class:`Model <htmd.model.Model>` object
        The model which to use to accumulate
    microvalue : ndarray
        An array of values corresponding to the microstates of the model

    Returns
    -------
    macrovalue : ndarray
        An array of values corresponding to the macrostates of the model
    """
    res = np.zeros(model.macronum)
    for i in range(len(microvalue)):
        macro = model.macro_ofmicro[i]
        res[macro] = res[macro] + microvalue[i]
    return res


def _sampleMacro(obj, macro, stConcat, mode, numFrames, replacement):
    from htmd.metricdata import _randomSample

    if mode == "random":
        frames = np.where(obj.macro_ofcluster[stConcat] == macro)[0]
        selFrames = _randomSample(frames, numFrames, replacement)
        selMicro = obj.micro_ofcluster[stConcat[selFrames]]
    elif mode == "even":
        micros = np.where(obj.macro_ofmicro == macro)[0]
        selFrames = []
        selMicro = []
        for i in range(len(micros)):
            selFrames.append(
                _sampleMicro(obj, micros[i], stConcat, numFrames, replacement)
            )
            selMicro.append(np.ones(numFrames) * micros[i])
    elif mode == "weighted":
        micros = np.where(obj.macro_ofmicro == macro)[0]
        eq = obj.msm.stationary_distribution
        weights = eq[micros] / np.sum(eq[micros])
        framespermicro = np.random.multinomial(numFrames, weights)
        selFrames = []
        selMicro = []
        for i in range(len(micros)):
            selFrames = np.append(
                selFrames,
                _sampleMicro(obj, micros[i], stConcat, framespermicro[i], replacement),
            )
            selMicro = np.append(selMicro, [micros[i]] * framespermicro[i])
    elif mode == "weightedTrunc":
        micros = np.where(obj.macro_ofmicro == macro)[0]
        eq = obj.msm.stationary_distribution
        weights = eq[micros] / np.sum(eq[micros])
        idx = np.argsort(weights)
        cs = np.cumsum(weights[idx])
        under50 = cs < 0.5
        weights[idx[under50]] = 0
        weights /= np.sum(weights)
        framespermicro = np.random.multinomial(numFrames, weights)
        selFrames = []
        selMicro = []
        for i in range(len(micros)):
            selFrames = np.append(
                selFrames,
                _sampleMicro(obj, micros[i], stConcat, framespermicro[i], replacement),
            )
            selMicro = np.append(selMicro, [micros[i]] * framespermicro[i])
    else:
        raise NameError("No valid mode given (read documentation)")
    return selFrames, selMicro


def _sampleMicro(obj, micro, stConcat, numFrames, replacement):
    from htmd.metricdata import _randomSample

    frames = np.where(obj.micro_ofcluster[stConcat] == micro)[0]
    return _randomSample(frames, numFrames, replacement)


def _macroTrajectoriesReport(macronum, macrost, simlist=None):
    """
    Prints out the number of trajectories that visited each macrostate and prints warnings

    Parameters
    ----------
    macronum
    macrost
    simlist
    """
    macrotraj = dict()
    macrotrajnum = np.zeros(macronum, dtype=int)
    for i, st in enumerate(macrost):
        macros = np.setdiff1d(np.unique(st), [-1])
        for m in macros:
            if m not in macrotraj:
                macrotraj[m] = []
            macrotraj[m].append(i)
            macrotrajnum[m] += 1
    logger.info("Number of trajectories that visited each macrostate:")
    logger.info(macrotrajnum)

    for m in range(macronum):
        ratio = macrotrajnum[m] / len(macrost)
        if macrotrajnum[m] <= 3 and ratio <= 0.2:
            logger.info(
                "Take care! Macro {} has been visited only in {} trajectories"
                " ({:.1f}% of total):".format(m, macrotrajnum[m], ratio * 100)
            )
            if simlist is not None:
                for s in macrotraj[m]:
                    logger.info(simlist[s])


def _macroTrajSt(St, macro_ofcluster):
    mst = []
    for i in range(len(St)):
        mst.append(macro_ofcluster[St[i]])
    return mst


"""def _macroP(C, macro_ofmicro):
    macronum = np.max(macro_ofmicro) + 1
    macroC = np.zeros((macronum, macronum))

    for m1 in range(macronum):
        sourcemicros = np.where(macro_ofmicro == m1)[0]
        for m2 in range(macronum):
            sinkmicros = np.where(macro_ofmicro == m2)[0]
            ixgrid = np.ix_(sourcemicros, sinkmicros)
            macroC[m1, m2] = np.sum(C[ixgrid].flatten())

    from deeptime.estimation import transition_matrix
    return transition_matrix(macroC, reversible=True)"""


class _TestModel(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        from htmd.simlist import simlist, simfilter
        from glob import glob
        from htmd.projections.metric import Metric
        from moleculekit.projections.metricdistance import MetricDistance
        from moleculekit.util import tempname
        from sklearn.cluster import MiniBatchKMeans
        from htmd.home import home
        from os.path import join

        sims = simlist(
            glob(join(home(dataDir="adaptive"), "data", "*", "")),
            glob(join(home(dataDir="adaptive"), "input", "*")),
        )
        fsims = simfilter(sims, tempname(), "not water")

        metr = Metric(fsims)
        metr.set(
            MetricDistance(
                "protein and resid 10 and name CA",
                "resname BEN and noh",
                periodic="selections",
                metric="contacts",
                groupsel1="residue",
                threshold=4,
            )
        )
        data = metr.project()
        data.cluster(MiniBatchKMeans(n_clusters=4))

        self.model = Model(data)

        self.trans_prob = np.array([[0.6, 0.2, 0.2], [0.3, 0.4, 0.3], [0.2, 0.3, 0.5]])

        self.trans_prob2 = np.array(
            [
                [0.8, 0.15, 0.05, 0.0, 0.0],
                [0.1, 0.75, 0.05, 0.05, 0.05],
                [0.05, 0.1, 0.8, 0.0, 0.05],
                [0.0, 0.2, 0.0, 0.8, 0.0],
                [1e-7, 0.02 - 1e-7, 0.02, 0.0, 0.96],
            ]
        )

    def test_model_saving_loading(self):
        from moleculekit.util import tempname

        modelfile = tempname(suffix=".dat")
        self.model.save(modelfile)

        newmodel = Model(file=modelfile)
        assert newmodel.data.numTrajectories == 2

        # Testing model saving when data has parent
        self.model.data.parent = self.model.data.copy()
        self.model.save(modelfile)
        newmodel = Model(file=modelfile)
        assert newmodel.data.numTrajectories == 2
        assert newmodel.data.parent.numTrajectories == 2
        self.model.data.parent = None

    def test_max_connected_lag(self):
        from htmd.metricdata import _generate_toy_data

        fakedata = _generate_toy_data(self.trans_prob, n_traj=100, seed=0)
        model = Model(fakedata)
        model.markovModel(1, 3)
        maxlag, its = model.maxConnectedLag([100, 500, 999], njobs=1)
        assert maxlag == 999
        assert np.allclose(
            its,
            np.array(
                [
                    [19.86198189, 15.92237368],
                    [81.7366474, 68.67389378],
                    [490.97275421, 330.52311174],
                ]
            ),
        )

    def test_plot_timescales(self):
        from htmd.metricdata import _generate_toy_data
        from matplotlib import pylab as plt
        import tempfile
        import os

        plt.ioff()
        fakedata = _generate_toy_data(self.trans_prob, n_traj=100, seed=0)
        model = Model(fakedata)

        with tempfile.TemporaryDirectory() as tmpdir:
            outplot = os.path.join(tmpdir, "test.png")
            its, lags = model.plotTimescales(
                lags=[1, 50, 100, 500, 800],
                plot=False,
                save=outplot,
                njobs=1,
                results=True,
            )
            assert os.path.exists(outplot)

        assert np.allclose(
            its,
            np.array(
                [
                    [0.98644281, 0.50313423],
                    [8.96353244, 8.72125828],
                    [19.86198189, 15.92237368],
                    [81.7366474, 68.67389378],
                    [183.99300253, 150.92253294],
                ]
            ),
        )
        assert np.array_equal(lags, np.array([1, 50, 100, 500, 800]))

        # Second probability matrix test
        fakedata = _generate_toy_data(self.trans_prob2, n_traj=100, seed=0)
        model = Model(fakedata)

        with tempfile.TemporaryDirectory() as tmpdir:
            outplot = os.path.join(tmpdir, "test.png")
            its, lags = model.plotTimescales(
                lags=range(1, 4),
                plot=False,
                save=outplot,
                njobs=1,
                results=True,
            )
            assert os.path.exists(outplot)

        assert np.allclose(
            its,
            np.array(
                [
                    [14.83314609, 4.8958021, 3.53366645, 2.01939226],
                    [14.82089692, 4.86804921, 3.49197211, 2.00651665],
                    [14.81046561, 4.8277952, 3.47217915, 2.02697621],
                ]
            ),
        )

    def test_ck_test(self):
        from htmd.metricdata import _generate_toy_data
        from matplotlib import pylab as plt
        import tempfile
        import os

        plt.ioff()
        fakedata = _generate_toy_data(self.trans_prob, n_traj=100, seed=0)
        model = Model(fakedata)
        model.markovModel(1, 3)
        with tempfile.TemporaryDirectory() as tmpdir:
            outplot = os.path.join(tmpdir, "test.png")
            model.cktest(plot=False, save=outplot)
            assert os.path.exists(outplot)

    def test_msm(self):
        from htmd.metricdata import _generate_toy_data

        fakedata = _generate_toy_data(self.trans_prob, n_traj=100, seed=0)
        model = Model(fakedata)
        model.markovModel(1, 3)

        assert np.allclose(self.trans_prob, model.P, atol=1e-1)

        eq_dist = np.array([0.28912558, 0.32754596, 0.38332846])
        assert np.allclose(eq_dist, model.eqDistribution(plot=False), atol=1e-3)

        assert np.array_equal(model.micro_ofcluster, [0, 1, 2])
        assert np.array_equal(model.cluster_ofmicro, [0, 1, 2])
        assert np.array_equal(model.macro_ofmicro, [2, 0, 1])
        assert model.micronum == 3
        assert model.macronum == 3

        # Second probability matrix test
        fakedata = _generate_toy_data(self.trans_prob2, n_traj=100, seed=0)
        model = Model(fakedata)
        model.markovModel(1, 3)

        assert np.allclose(self.trans_prob2, model.P, atol=1e-1)

        eq_dist = np.array([0.14541275, 0.33704423, 0.51754301])
        assert np.allclose(eq_dist, model.eqDistribution(plot=False), atol=1e-3)

        assert np.array_equal(model.micro_ofcluster, [0, 1, 2, 3, 4])
        assert np.array_equal(model.cluster_ofmicro, [0, 1, 2, 3, 4])
        assert np.array_equal(model.macro_ofmicro, [1, 1, 1, 0, 2])
        assert model.micronum == 5
        assert model.macronum == 3

    def test_create_state(self):
        from htmd.metricdata import _generate_toy_data

        fakedata = _generate_toy_data(self.trans_prob, n_traj=100, seed=0)
        model = Model(fakedata)
        model.markovModel(1, 3)
        # Create a new macrostate assigning microstates 0 and 1 to it
        model.createState(microstates=[0, 1])
        assert model.macronum == 2
        assert np.array_equal(model.macro_ofmicro, [1, 1, 0])
        model.markovModel(1, 3)  # Revert changes

        # Create a new cluster out of the first frames of the first trajectory
        model.createState(indexpairs=[[0, 0], [0, 5], [0, 10]])
        model.markovModel(1, 3)
        assert np.array_equal(model.micro_ofcluster, [0, 1, 2, 3])
        assert np.array_equal(model.cluster_ofmicro, [0, 1, 2, 3])
        assert np.array_equal(model.macro_ofmicro, [2, 0, 1, 2])

        # Create a new cluster out of the first frames of the first trajectory without back transitions
        fakedata = _generate_toy_data(self.trans_prob, n_traj=100, seed=0)
        model = Model(fakedata)
        model.markovModel(1, 3)
        model.createState(indexpairs=[[0, 0], [0, 1], [0, 2]])  # No back transitions
        assert np.array_equal(model.micro_ofcluster, [0, 1, 2, -1])
        assert np.array_equal(model.cluster_ofmicro, [0, 1, 2])
        assert np.array_equal(model.macro_ofmicro, [2, 0, 1])

    def test_core_set_model(self):
        from htmd.metricdata import _generate_toy_data

        fakedata = _generate_toy_data(self.trans_prob, n_traj=100, seed=0)
        model = Model(fakedata)
        model.createState(indexpairs=[[0, 0], [0, 5], [0, 10]])  # Create a rare state
        model.markovModel(1, 3)
        assert np.array_equal(model.micro_ofcluster, [0, 1, 2, 3])
        assert np.array_equal(model.cluster_ofmicro, [0, 1, 2, 3])
        assert np.array_equal(model.macro_ofmicro, [2, 0, 1, 2])

        coremodel, _ = model.createCoreSetModel(threshold=0.5)
        coremodel.markovModel(1, 3)
        assert np.array_equal(coremodel.micro_ofcluster, [0, 1, 2])
        assert np.array_equal(coremodel.cluster_ofmicro, [0, 1, 2])
        assert np.array_equal(coremodel.macro_ofmicro, [0, 1, 2])

    def test_plot_fes(self):
        from htmd.metricdata import _generate_toy_data
        from matplotlib import pylab as plt
        import tempfile
        import os

        plt.ioff()

        fakedata = _generate_toy_data(self.trans_prob, n_traj=100, seed=0)
        model = Model(fakedata)
        model.markovModel(1, 3)

        with tempfile.TemporaryDirectory() as tmpdir:
            outplot = os.path.join(tmpdir, "test.png")
            model.plotFES(0, 1, 300, states=True, s=10, plot=False, save=outplot)
            assert os.path.exists(outplot)

        fakedata2 = _generate_toy_data(
            self.trans_prob, n_traj=100, seed=0, cluster=False
        )
        # Plot with second data object
        with tempfile.TemporaryDirectory() as tmpdir:
            outplot = os.path.join(tmpdir, "test.png")
            model.plotFES(
                0, 1, 300, states=True, s=10, plot=False, data=fakedata2, save=outplot
            )
            assert os.path.exists(outplot)

    def test_sample_states(self):
        from htmd.metricdata import _generate_toy_data

        fakedata = _generate_toy_data(self.trans_prob, n_traj=100, seed=0)
        model = Model(fakedata)
        model.markovModel(1, 3)
        np.random.seed(0)
        model.sampleStates(states=[0, 1])

    def test_get_state_statistic(self):
        from htmd.metricdata import _generate_toy_data

        fakedata = _generate_toy_data(self.trans_prob, n_traj=100, seed=0)
        model = Model(fakedata)
        model.markovModel(1, 3)
        statestats = getStateStatistic(
            model, fakedata, states=[0, 1], statetype="micro", method=np.mean
        )
        assert np.allclose(statestats[0], [1, 0, 0])
        assert np.allclose(statestats[1], [0, 1, 0])

        statestats = getStateStatistic(
            model, fakedata, states=[0, 1, 2], statetype="cluster", method=np.mean
        )
        assert np.allclose(statestats[0], [1, 0, 0])
        assert np.allclose(statestats[1], [0, 1, 0])
        assert np.allclose(statestats[2], [0, 0, 1])

        model.createState(microstates=[0, 1])
        statestats = getStateStatistic(
            model, fakedata, states=[0, 1], statetype="macro", method=np.mean
        )
        assert np.allclose(statestats[0], [0, 0, 1])
        assert np.allclose(statestats[1], [0.56985649, 0.43014351, 0])

    def test_macro_accumulate(self):
        from htmd.metricdata import _generate_toy_data

        fakedata = _generate_toy_data(self.trans_prob, n_traj=100, seed=0)
        model = Model(fakedata)
        model.markovModel(1, 3)
        acc = macroAccumulate(model, [1.5, 3.7, 4.5])
        assert np.array_equal(acc, np.array([3.7, 4.5, 1.5]))

        model.createState(microstates=[0, 1])
        acc = macroAccumulate(model, [1.5, 3.7, 4.5])
        assert np.array_equal(acc, np.array([4.5, 5.2]))


if __name__ == "__main__":
    unittest.main(verbosity=2)
